# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019, 2020 Carlo Pavan <carlopav@gmail.com>             *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provide the Draft_Edit command used by the Draft workbench."""
## @package gui_edit
# \ingroup DRAFT
# \brief Provide the Draft_Edit command used by the Draft workbench

import math
from pivy import coin
from PySide import QtCore, QtGui

import FreeCAD as App
import FreeCADGui as Gui
import Draft
import DraftTools
from draftutils.translate import translate
import draftguitools.gui_trackers as trackers

__title__ = "FreeCAD Draft Edit Tool"
__author__ = ("Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, "
              "Dmitry Chigrin, Carlo Pavan")
__url__ = "https://www.freecadweb.org"

COLORS = {
    "default": Gui.draftToolBar.getDefaultColor("snap"),
    "black":  (0., 0., 0.),
    "white":  (1., 1., 1.),
    "grey":   (.5, .5, .5),
    "red":    (1., 0., 0.),
    "green":  (0., 1., 0.),
    "blue":   (0., 0., 1.),
    "yellow": (1., 1., 0.),
    "cyan":   (0., 1., 1.),
    "magenta": (1., 0., 1.)
}


class Edit:
    """The Draft_Edit FreeCAD command definition.

    A tool to graphically edit FreeCAD objects.
    Current implementation use many parts of pivy graphics code by user "looo".
    The tool collect editpoints from objects and display Trackers on them
    to allow editing their Shape and their parameters.

    Callbacks
    ---------
    selection_callback
        registered when tool is launched, identify
        selected objects.

    editing_callbacks
        self._keyPressedCB      -> self.keyPressed
        self._mouseMovedCB      -> self._mouseMovedCB
        if self._mousePressedCB -> self.mousePressed
        when trackers are displayed for selected objects,
        these callbacks capture user events and forward
        them to related functions

    Task panel (Draft Toolbar)
    ----------
    self.ui = Gui.draftToolBar
    TODO: since we introduced context menu for interacting
          with editTrackers, point 2 should become obsolete,
          because not consistent with multi-object editing.

    Draft_Edit uses taskpanel in 3 ways:

    1 - when waiting for user to select an object
        calling self.ui.selectUi()

    2 - when Trackers are displayed and user must click one, a
        custom task panel is displayed depending on edited
        object:
        self.ui.editUi()            -> the default one
        self.ui.editUi("Wire")      -> line and wire editing
        self.ui.editUi("BezCurve")  -> BezCurve editing
        self.ui.editUi("Circle")    -> circle editing
        self.ui.editUi("Arc")       -> arc editing
        When Draft_Edit evaluate mouse click, depending if some
        ui button have been pressed (.isChecked()), decide if
        the action is a startEditing or AddPoint or DelPoint or
        change BezCurve Continuity, ecc.

    3 - when in editing, lineUi support clicking destination point
        by self.startEditing
        self.ui.lineUi()
        self.ui.isRelative.show()

    Tracker selection
    -----------------
    If the tool recognize mouse click as an attempt to startEditing,
    using soRayPickAction, it identifies the selected editTracker and
    start editing it. Here is where "looo" code was very useful.


    Editing preview
    ---------------
    When object editing begins, self.ghost is initiated with the
    corresponding DraftTracker of the object type. The object Tracker
    is deleted when user clicks again and endEditing.

    Context Menu
    ------------
    Activated with Alt+LeftClick or pressing key "e"
    It's a custom context menu, that depends on clicked tracker
    or on clicked object.

    display_tracker_menu
        populates the menu with custom actions

    evaluate_menu_action
        evaluate user chosen action and launch corresponding
        function.

    Preferences
    -----------
    maxObjects: Int
        set by "DraftEditMaxObjects" in user preferences
        The max number of FreeCAD objects the tool is
        allowed to edit at the same time.

    pick_radius: Int
        set by "DraftEditPickRadius" in user preferences
        The pick radius during editing operation.
        Increase if you experience problems in clicking
        on a editTracker because of screen resolution.

    Attributes
    ----------
    obj: Edited object
        I'm planning to discard this attribute.
        In old implementation every function was supposed to
        act on self.obj, self.editpoints, self.trackers,
        self.pl, self.invpl.
        Due to multiple object editing, i'm planning to keep
        just self.trackers. Any other object will be identified
        and processed starting from editTracker information.

    editing: Int
        Index of the editTracker that has been clicked by the
        user. Tracker selection mechanism is based on it.
        if self.editing is None :
            the user didn't click any node, and next click will
            be processed as an attempt to start editing operation
        if self.editing == o or 1 or 2 or 3 etc :
            the user is editing corresponding node, so next click
            will be processed as an attempt to end editing operation

    editpoints: List [FreeCAD::App.Vector]
        List of editpoints collected from the edited object,
        on whick editTrackers will be placed.

    trackers: Dictionary {object.Name : [editTrackers]}
        It records the list of DraftTrackers.editTracker.
        {object.Name as String : [editTrackers for the object]}
        Each tracker is created with (position,obj.Name,idx),
        so it's possible to recall it
        self.trackers[str(node.objectName.getValue())][ep]

    overNode: DraftTrackers.editTracker
        It represent the editTracker under the cursor position.
        It is used to preview the tracker selection action.

    ghost: DraftTrackers.*
        Handles the tracker to preview editing operations.
        it is initialized when user clicks on a editTracker
        by self.startEditing() function.

    alt_edit_mode: Int
        Allows alternative editing modes for objects.
        ATM supported for:
        - arcs: if 0 edit by 3 points, if 1 edit by center,
                radius, angles

    supportedObjs: List
        List of supported Draft Objects.
        The tool use Draft.getType(obj) to compare object type
        to the list.

    supportedPartObjs: List
        List of supported Part Objects.
        The tool use Draft.getType(obj) and obj.TypeId to compare
        object type to the list.
    """

    def __init__(self):
        """Initialize Draft_Edit Command."""
        self.running = False
        self.trackers = {'object': []}
        self.overNode = None  # preselected node with mouseover
        self.obj = None
        self.editing = None

        # event callbacks
        self.selection_callback = None
        self._keyPressedCB = None
        self._mouseMovedCB = None
        self._mousePressedCB = None

        # this are used to edit structure objects, it's a bit buggy i think
        self.selectstate = None
        self.originalDisplayMode = None
        self.originalPoints = None
        self.originalNodes = None

        # settings
        param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
        self.maxObjects = param.GetInt("DraftEditMaxObjects", 5)
        self.pick_radius = param.GetInt("DraftEditPickRadius", 20)
        
        self.alt_edit_mode = 0 # alternative edit mode for objects

        # preview
        self.ghost = None

        #list of supported Draft and Arch objects
        self.supportedObjs = ["BezCurve","Wire","BSpline","Circle","Rectangle",
                            "Polygon","Dimension","Space","Structure","PanelCut",
                            "PanelSheet","Wall", "Window"]

        #list of supported Part objects (they don't have a proxy)
        #TODO: Add support for "Part::Circle" "Part::RegularPolygon" "Part::Plane" "Part::Ellipse" "Part::Vertex" "Part::Spiral"
        self.supportedPartObjs = ["Sketch", "Sketcher::SketchObject",
                                "Part", "Part::Line", "Part::Box"]

    def GetResources(self):
        
        tooltip = ("Edits the active object.\n"
                   "Press E or ALT+LeftClick to display context menu\n"
                   "on supported nodes and on supported objects.")
                  
        return {'Pixmap': 'Draft_Edit',
                'Accel': "D, E",
                'MenuText': QtCore.QT_TRANSLATE_NOOP("Draft_Edit", "Edit"),
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("Draft_Edit", tooltip)
                }


    # -------------------------------------------------------------------------
    # MAIN FUNCTIONS
    # -------------------------------------------------------------------------

    def Activated(self):
        """
        Activated is run when user launch Edit command.
        If something is selected -> call self.proceed()
        If nothing is selected -> self.register_selection_callback()
        """
        if self.running:
            self.finish()
        DraftTools.Modifier.Activated(self, "Edit")
        if not App.ActiveDocument:
            self.finish()

        self.ui = Gui.draftToolBar
        self.view = Draft.get3DView()

        if Gui.Selection.getSelection():
            self.proceed()
        else:
            self.ui.selectUi()
            App.Console.PrintMessage(translate("draft", 
                                               "Select a Draft object to edit")
                                               + "\n")
            self.register_selection_callback()

    def proceed(self):
        """this method defines editpoints and set the editTrackers"""
        self.unregister_selection_callback()
        self.edited_objects = self.getObjsFromSelection()
        if not self.edited_objects:
            return self.finish()

        # Save selectstate and turn selectable false.
        # Object can remain selectable commenting following lines:
        # self.saveSelectState(self.obj)
        # self.setSelectState(self.obj, False)

        # start object editing
        Gui.Selection.clearSelection()
        Gui.Snapper.setSelectMode(True)

        self.ui.editUi()

        for obj in self.edited_objects:
            self.setEditPoints(obj)

        self.register_editing_callbacks()

        # TODO: align working plane when editing starts
        # App.DraftWorkingPlane.save()
        # self.alignWorkingPlane()


    def finish(self, closed=False):
        """Terminate Edit Tool."""
        self.unregister_selection_callback()
        self.unregister_editing_callbacks()
        self.editing = None
        self.finalizeGhost()
        Gui.Snapper.setSelectMode(False)
        if self.obj and closed:
            if "Closed" in self.obj.PropertiesList:
                if not self.obj.Closed:
                    self.obj.Closed = True
        if self.ui:
            self.removeTrackers()
        self.restoreSelectState(self.obj)
        if Draft.getType(self.obj) == "Structure":
            if self.originalDisplayMode is not None:
                self.obj.ViewObject.DisplayMode = self.originalDisplayMode
            if self.originalPoints is not None:
                self.obj.ViewObject.NodeSize = self.originalPoints
            if self.originalNodes is not None:
                self.obj.ViewObject.ShowNodes = self.originalNodes
        self.selectstate = None
        self.originalDisplayMode = None
        self.originalPoints = None
        self.originalNodes = None
        DraftTools.Modifier.finish(self)
        App.DraftWorkingPlane.restore()
        if Gui.Snapper.grid:
            Gui.Snapper.grid.set()
        self.running = False
        # delay resetting edit mode otherwise it doesn't happen
        from PySide import QtCore
        QtCore.QTimer.singleShot(0, Gui.ActiveDocument.resetEdit)

    # -------------------------------------------------------------------------
    # SCENE EVENTS CALLBACKS
    # -------------------------------------------------------------------------

    def register_selection_callback(self):
        """Register callback for selection when command is launched."""
        self.unregister_selection_callback()
        self.selection_callback = self.view.addEventCallback("SoEvent",DraftTools.selectObject)

    def unregister_selection_callback(self):
        """
        remove selection callback if it exists
        """
        if self.selection_callback:
            self.view.removeEventCallback("SoEvent", self.selection_callback)
        self.selection_callback = None

    def register_editing_callbacks(self):
        """
        register editing callbacks (former action function)
        """
        viewer = Gui.ActiveDocument.ActiveView.getViewer()
        self.render_manager = viewer.getSoRenderManager()
        view = Gui.ActiveDocument.ActiveView
        if self._keyPressedCB is None:
            self._keyPressedCB = view.addEventCallbackPivy(
            coin.SoKeyboardEvent.getClassTypeId(), self.keyPressed)
        if self._mouseMovedCB is None:
            self._mouseMovedCB = view.addEventCallbackPivy(
            coin.SoLocation2Event.getClassTypeId(), self.mouseMoved)
        if self._mousePressedCB is None:
            self._mousePressedCB = view.addEventCallbackPivy(
            coin.SoMouseButtonEvent.getClassTypeId(), self.mousePressed)
        #App.Console.PrintMessage("Draft edit callbacks registered \n")

    def unregister_editing_callbacks(self):
        """
        remove callbacks used during editing if they exist
        """
        view = Gui.ActiveDocument.ActiveView
        if self._keyPressedCB:
            view.removeEventCallbackSWIG(coin.SoKeyboardEvent.getClassTypeId(), self._keyPressedCB)
            self._keyPressedCB = None
            #App.Console.PrintMessage("Draft edit keyboard callback unregistered \n")
        if self._mouseMovedCB:
            view.removeEventCallbackSWIG(coin.SoLocation2Event.getClassTypeId(), self._mouseMovedCB)
            self._mouseMovedCB = None
            #App.Console.PrintMessage("Draft edit location callback unregistered \n")
        if self._mousePressedCB:
            view.removeEventCallbackSWIG(coin.SoMouseButtonEvent.getClassTypeId(), self._mousePressedCB)
            self._mousePressedCB = None
            #App.Console.PrintMessage("Draft edit mouse button callback unregistered \n")

    # -------------------------------------------------------------------------
    # SCENE EVENT HANDLERS
    # -------------------------------------------------------------------------

    def keyPressed(self, event_callback):
        """Execute as callback for keyboard event."""
        # TODO: Get the keys from preferences
        event = event_callback.getEvent()
        if event.getState() == coin.SoKeyboardEvent.DOWN:
            key = event.getKey()
            # App.Console.PrintMessage("pressed key : "+str(key)+"\n")
            if key == 65307:  # ESC
                self.finish()
            if key == 97:  # "a"
                self.finish()
            if key == 111:  # "o"
                self.finish(closed=True)
            if key == 101:  # "e"
                self.display_tracker_menu(event)
            if key == 105:  # "i"
                if Draft.getType(self.obj) == "Circle":
                    self.arcInvert(self.obj)

    def mousePressed(self, event_callback):
        """
        mouse button event handler, calls: startEditing, endEditing, addPoint, delPoint
        """
        event = event_callback.getEvent()
        if (event.getState() == coin.SoMouseButtonEvent.DOWN and 
            event.getButton() == event.BUTTON1
            ):#left click
            if not event.wasAltDown():
                if self.editing is None:
                    self.startEditing(event)
                else:
                    self.endEditing(self.obj, self.editing)
            elif event.wasAltDown():  # left click with ctrl down
                self.display_tracker_menu(event)

    def mouseMoved(self, event_callback):
        """Execute as callback for mouse movement.

        Update tracker position and update preview ghost.
        """
        event = event_callback.getEvent()
        if self.editing is not None:
            self.updateTrackerAndGhost(event)
        else:
            # look for a node in mouse position and highlight it
            pos = event.getPosition()
            node = self.getEditNode(pos)
            ep = self.getEditNodeIndex(node)
            if ep is not None:
                if self.overNode is not None:
                    self.overNode.setColor(COLORS["default"])
                self.trackers[str(node.objectName.getValue())][ep].setColor(COLORS["red"])
                self.overNode = self.trackers[str(node.objectName.getValue())][ep]
            else:
                if self.overNode is not None:
                    self.overNode.setColor(COLORS["default"])
                    self.overNode = None

    def startEditing(self, event):
        """Start editing selected EditNode."""
        pos = event.getPosition()
        node = self.getEditNode(pos)
        ep = self.getEditNodeIndex(node)
        if ep is None:
            return

        doc = App.getDocument(str(node.documentName.getValue()))
        self.obj = doc.getObject(str(node.objectName.getValue()))
        if self.obj is None:
            return
        self.setPlacement(self.obj)

        App.Console.PrintMessage(self.obj.Name
                                 + ": editing node number "
                                 + str(ep) + "\n")

        self.ui.lineUi()
        self.ui.isRelative.show()
        self.editing = ep
        self.trackers[self.obj.Name][self.editing].off()
        self.finalizeGhost()
        self.ghost = self.initGhost(self.obj)
        self.node.append(self.trackers[self.obj.Name][self.editing].get())
        Gui.Snapper.setSelectMode(False)
        self.hideTrackers()

    def updateTrackerAndGhost(self, event):
        """Update tracker position when editing and update ghost."""
        pos = event.getPosition().getValue()
        orthoConstrain = False
        if event.wasShiftDown() == 1:
            orthoConstrain = True
        snappedPos = Gui.Snapper.snap((pos[0],pos[1]),self.node[-1], constrain=orthoConstrain)
        self.trackers[self.obj.Name][self.editing].set(snappedPos)
        self.ui.displayPoint(snappedPos, self.node[-1])
        if self.ghost:
            self.updateGhost(obj=self.obj, idx=self.editing, pt=snappedPos)

    def endEditing(self, obj, nodeIndex, v=None):
        """Terminate editing and start object updating process."""
        self.finalizeGhost()
        self.trackers[obj.Name][nodeIndex].on()
        Gui.Snapper.setSelectMode(True)
        if v is None:
            # endEditing is called by mousePressed
            v = self.trackers[obj.Name][nodeIndex].get()
        else:
            # endEditing is called by numericInput, so tracker
            # position should be updated manually
            self.trackers[obj.Name][nodeIndex].set(v)
        self.update(obj, nodeIndex, v)
        self.alt_edit_mode = 0
        self.ui.editUi(self.ui.lastMode)
        self.node = []
        self.editing = None
        self.showTrackers()
        DraftTools.redraw3DView()

    # -------------------------------------------------------------------------
    # UTILS
    # -------------------------------------------------------------------------

    def getObjsFromSelection(self):
        """Evaluate selection and return a valid object to edit."""
        selection = Gui.Selection.getSelection()
        self.edited_objects = []
        if len(selection) > self.maxObjects:
            App.Console.PrintMessage(translate("draft", 
                                               "Too many objects selected, max number set to: ")
                                     + str(self.maxObjects) + "\n")
            return None
        for obj in selection:
            if Draft.getType(obj) in self.supportedObjs:
                self.edited_objects.append(obj)
                continue
            elif Draft.getType(obj) in self.supportedPartObjs:
                if obj.TypeId in self.supportedPartObjs:
                    self.edited_objects.append(obj)
                    continue
            App.Console.PrintWarning(obj.Name 
                                     + translate("draft",
                                                 ": this object is not editable")
                                     + "\n")
        return self.edited_objects

    def get_selected_obj_at_position(self, pos):
        """Return object at given position.

        If object is one of the edited objects (self.edited_objects).
        """
        selobjs = Gui.ActiveDocument.ActiveView.getObjectsInfo((pos[0],pos[1]))
        if not selobjs:
            return
        for info in selobjs:
            if not info:
                return
            for obj in self.edited_objects:
                if obj.Name == info["Object"]:
                    return obj

    def numericInput(self, v, numy=None, numz=None):
        """Execute callback by the toolbar to activate the update function.

        This function gets called by the toolbar
        or by the mouse click and activate the update function.
        """
        if numy:
            v = App.Vector(v, numy, numz)
        self.endEditing(self.obj, self.editing, v)
        App.ActiveDocument.recompute()

    def setSelectState(self, obj, selState=False):
        if hasattr(obj.ViewObject, "Selectable"):
            obj.ViewObject.Selectable = selState

    def saveSelectState(self, obj):
        if hasattr(obj.ViewObject, "Selectable"):
            self.selectstate = obj.ViewObject.Selectable

    def restoreSelectState(self,obj):
        if obj:
            if hasattr(obj.ViewObject,"Selectable") and (self.selectstate is not None):
                obj.ViewObject.Selectable = self.selectstate

    def setPlacement(self, obj):
        """Set placement of object.

        Set self.pl and self.invpl to self.obj placement
        and inverse placement.
        """
        if not obj:
            return
        if "Placement" in obj.PropertiesList:
            self.pl = obj.getGlobalPlacement()
            self.invpl = self.pl.inverse()

    def alignWorkingPlane(self):
        """Align working plane to self.obj."""
        if "Shape" in self.obj.PropertiesList:
            if DraftTools.plane.weak:
                DraftTools.plane.alignToFace(self.obj.Shape)
        if self.planetrack:
            self.planetrack.set(self.editpoints[0])

    def getEditNode(self, pos):
        """Get edit node from given screen position."""
        node = self.sendRay(pos)
        return node
    
    def sendRay(self, mouse_pos):
        """Send a ray through the scene and return the nearest entity."""
        ray_pick = coin.SoRayPickAction(self.render_manager.getViewportRegion())
        ray_pick.setPoint(coin.SbVec2s(*mouse_pos))
        ray_pick.setRadius(self.pick_radius)
        ray_pick.setPickAll(True)
        ray_pick.apply(self.render_manager.getSceneGraph())
        picked_point = ray_pick.getPickedPointList()
        return self.searchEditNode(picked_point)

    def searchEditNode(self, picked_point):
        """Search edit node inside picked point list and return node number."""
        for point in picked_point:
            path = point.getPath()
            length = path.getLength()
            point = path.getNode(length - 2)
            #import DraftTrackers
            if hasattr(point,"subElementName") and 'EditNode' in str(point.subElementName.getValue()):
                return point
        return None

    def getEditNodeIndex(self, point):
        """Get edit node index from given screen position."""
        if point:
            subElement = str(point.subElementName.getValue())
            ep = int(subElement[8:])
            return ep
        else:
            return None

    # -------------------------------------------------------------------------
    # EDIT TRACKERS functions
    # -------------------------------------------------------------------------

    def setTrackers(self, obj, points=None):
        """Set Edit Trackers for editpoints collected from self.obj."""
        if points is None or len(points) == 0:
            App.Console.PrintWarning(translate("draft", 
                                               "No edit point found for selected object")
                                               + "\n")
            # do not finish if some trackers are still present
            if self.trackers == {'object': []}:
                self.finish()
            return
        self.trackers[obj.Name] = []
        if Draft.getType(obj) == "BezCurve":
            self.resetTrackersBezier(obj)
        else:
            if obj.Name in self.trackers:
                self.removeTrackers(obj)
            for ep in range(len(points)):
                self.trackers[obj.Name].append(trackers.editTracker(pos=points[ep],name=obj.Name,idx=ep))

    def resetTrackers(self, obj):
        """Reset Edit Trackers and set them again."""
        self.removeTrackers(obj)
        self.setTrackers(obj, self.getEditPoints(obj))

    def removeTrackers(self, obj=None):
        """Remove Edit Trackers."""
        if obj:
            if obj.Name in self.trackers:
                for t in self.trackers[obj.Name]:
                    t.finalize()
            self.trackers[obj.Name] = []
        else:
            for key in self.trackers.keys():
                for t in self.trackers[key]:
                    t.finalize()
            self.trackers = {'object': []}


    def hideTrackers(self, obj=None):
        """Hide Edit Trackers.

        Attributes
        ----------
        obj: FreeCAD object
            Hides trackers only for given object,
            if obj is None, hides all trackers
        """
        if obj is None:
            for key in self.trackers:
                for t in self.trackers[key]:
                    t.off()
        else:
            for t in self.trackers[obj.Name]:
                t.off()

    def showTrackers(self, obj=None):
        """Show Edit Trackers.

        Attributes
        ----------
        obj: FreeCAD object
            Shows trackers only for given object,
            if obj is None, shows all trackers
        """
        if obj is None:
            for key in self.trackers:
                for t in self.trackers[key]:
                    t.on()
        else:
            for t in self.trackers[obj.Name]:
                t.on()

    # -------------------------------------------------------------------------
    # PREVIEW
    # -------------------------------------------------------------------------

    def initGhost(self, obj):
        """Initialize preview ghost."""
        if Draft.getType(obj) == "Wire":
            return trackers.wireTracker(obj.Shape)
        elif Draft.getType(obj) == "BSpline":
            return trackers.bsplineTracker()
        elif Draft.getType(obj) == "BezCurve":
            return trackers.bezcurveTracker()
        elif Draft.getType(obj) == "Circle":
            return trackers.arcTracker()

    def updateGhost(self, obj, idx, pt):
        if Draft.getType(obj) in ["Wire"]:
            self.ghost.on()
            pointList = self.applyPlacement(obj.Points)
            pointList[idx] = pt
            if obj.Closed:
                pointList.append(pointList[0])
            self.ghost.updateFromPointlist(pointList)
        elif Draft.getType(obj) == "BSpline":
            self.ghost.on()
            pointList = self.applyPlacement(obj.Points)
            pointList[idx] = pt
            if obj.Closed:
                pointList.append(pointList[0])
            self.ghost.update(pointList)
        elif Draft.getType(obj) == "BezCurve":
            self.ghost.on()
            plist = self.applyPlacement(obj.Points)
            pointList = self.recomputePointsBezier(obj,plist,idx,pt,obj.Degree,moveTrackers=True)
            self.ghost.update(pointList,obj.Degree)
        elif Draft.getType(obj) == "Circle":
            self.ghost.on()
            self.ghost.setCenter(obj.getGlobalPlacement().Base)
            self.ghost.setRadius(obj.Radius)
            if self.obj.FirstAngle == self.obj.LastAngle:
                # self.obj is a circle
                self.ghost.circle = True
                if self.editing == 0:
                    self.ghost.setCenter(pt)
                elif self.editing == 1:
                    radius = pt.sub(obj.getGlobalPlacement().Base).Length
                    self.ghost.setRadius(radius)
            else:
                if self.alt_edit_mode == 0:
                    # edit by 3 points
                    if self.editing == 0:
                        # center point
                        import DraftVecUtils
                        p1 = self.invpl.multVec(self.obj.Shape.Vertexes[0].Point)
                        p2 = self.invpl.multVec(self.obj.Shape.Vertexes[1].Point)
                        p0 = DraftVecUtils.project(self.invpl.multVec(pt),self.invpl.multVec(self.getArcMid(obj, global_placement=True)))
                        self.ghost.autoinvert=False
                        self.ghost.setRadius(p1.sub(p0).Length)
                        self.ghost.setStartPoint(self.obj.Shape.Vertexes[1].Point)
                        self.ghost.setEndPoint(self.obj.Shape.Vertexes[0].Point)
                        self.ghost.setCenter(self.pl.multVec(p0))
                        return
                    else:
                        p1 = self.getArcStart(obj, global_placement=True)
                        p2 = self.getArcMid(obj, global_placement=True)
                        p3 = self.getArcEnd(obj, global_placement=True)
                        if self.editing == 1:
                            p1=pt
                        elif self.editing == 3:
                            p2=pt
                        elif self.editing == 2:
                            p3=pt
                        self.ghost.setBy3Points(p1,p2,p3)
                elif self.alt_edit_mode == 1:
                    # edit by center radius angles
                    self.ghost.setStartAngle(math.radians(obj.FirstAngle))
                    self.ghost.setEndAngle(math.radians(obj.LastAngle))
                    if self.editing == 0:
                        self.ghost.setCenter(pt)
                    elif self.editing == 1:
                        self.ghost.setStartPoint(pt)
                    elif self.editing == 2:
                        self.ghost.setEndPoint(pt)
                    elif self.editing == 3:
                        self.ghost.setRadius(self.invpl.multVec(pt).Length)
        DraftTools.redraw3DView()

    def applyPlacement(self, pointList):
        if self.pl:
            plist = []
            for p in pointList:
                point = self.pl.multVec(p)
                plist.append(point)
            return plist
        else:
            return pointList

    def finalizeGhost(self):
        try:
            self.ghost.finalize()
            self.ghost = None
        except:
            return

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Add/Delete Vertexes
    # -------------------------------------------------------------------------

    def addPoint(self, event):
        """Execute callback, add point to obj and reset trackers."""
        pos = event.getPosition()
        # self.setSelectState(self.obj, True)
        selobjs = Gui.ActiveDocument.ActiveView.getObjectsInfo((pos[0],pos[1]))
        if not selobjs:
            return
        for info in selobjs:
            if not info:
                return
            for o in self.edited_objects:
                if o.Name != info["Object"]:
                    continue
                self.obj = o
                break
            self.setPlacement(self.obj)
            if Draft.getType(self.obj) == "Wire" and 'Edge' in info["Component"]:
                pt = App.Vector(info["x"], info["y"], info["z"])
                self.addPointToWire(self.obj, pt, int(info["Component"][4:]))
            elif Draft.getType(self.obj) in ["BSpline", "BezCurve"]: #to fix double vertex created
                # pt = self.point
                if "x" in info:# prefer "real" 3D location over working-plane-driven one if possible
                    pt = App.Vector(info["x"], info["y"], info["z"])
                else:
                    continue
                self.addPointToCurve(pt, info)
        self.obj.recompute()
        self.removeTrackers(self.obj)
        self.setEditPoints(self.obj)
        # self.setSelectState(self.obj, False)
        return

    def addPointToWire(self, obj, newPoint, edgeIndex):
        newPoints = []
        hasAddedPoint = False
        if hasattr(obj, "ChamferSize") and hasattr(obj, "FilletRadius"):
            if obj.ChamferSize > 0 and obj.FilletRadius > 0:
                edgeIndex = (edgeIndex + 3) / 4
            elif obj.ChamferSize > 0 or obj.FilletRadius > 0:
                edgeIndex = (edgeIndex + 1) / 2

        for index, point in enumerate(self.obj.Points):
            if index == edgeIndex:
                hasAddedPoint = True
                newPoints.append(self.invpl.multVec(newPoint))
            newPoints.append(point)
        if obj.Closed and edgeIndex == len(obj.Points):
            # last segment when object is closed
            newPoints.append(self.invpl.multVec(newPoint))
        obj.Points = newPoints

    def addPointToCurve(self, point, info=None):
        import Part
        if not (Draft.getType(self.obj) in ["BSpline", "BezCurve"]):
            return
        pts = self.obj.Points
        if Draft.getType(self.obj) == "BezCurve":
            if not info['Component'].startswith('Edge'):
                return  # clicked control point
            edgeindex = int(info['Component'].lstrip('Edge')) - 1
            wire = self.obj.Shape.Wires[0]
            bz = wire.Edges[edgeindex].Curve
            param = bz.parameter(point)
            seg1 = wire.Edges[edgeindex].copy().Curve
            seg2 = wire.Edges[edgeindex].copy().Curve
            seg1.segment(seg1.FirstParameter, param)
            seg2.segment(param, seg2.LastParameter)
            if edgeindex == len(wire.Edges):
                # we hit the last segment, we need to fix the degree
                degree=wire.Edges[0].Curve.Degree
                seg1.increase(degree)
                seg2.increase(degree)
            edges = wire.Edges[0:edgeindex] + [Part.Edge(seg1),Part.Edge(seg2)] \
                + wire.Edges[edgeindex + 1:]
            pts = edges[0].Curve.getPoles()[0:1]
            for edge in edges:
                pts.extend(edge.Curve.getPoles()[1:])
            if self.obj.Closed:
                pts.pop()
            c = self.obj.Continuity
            # assume we have a tangent continuity for an arbitrarily split
            # segment, unless it's linear
            cont = 1 if (self.obj.Degree >= 2) else 0
            self.obj.Continuity = c[0:edgeindex] + [cont] + c[edgeindex:]
        else:
            if (Draft.getType(self.obj) in ["BSpline"]):
                if (self.obj.Closed == True):
                    curve = self.obj.Shape.Edges[0].Curve
                else:
                    curve = self.obj.Shape.Curve
            uNewPoint = curve.parameter(point)
            uPoints = []
            for p in self.obj.Points:
                uPoints.append(curve.parameter(p))
            for i in range(len(uPoints) - 1):
                if ( uNewPoint > uPoints[i] ) and ( uNewPoint < uPoints[i+1] ):
                    pts.insert(i + 1, self.invpl.multVec(point))
                    break
            # DNC: fix: add points to last segment if curve is closed
            if self.obj.Closed and (uNewPoint > uPoints[-1]):
                pts.append(self.invpl.multVec(point))
        self.obj.Points = pts

    def delPoint(self, event):
        pos = event.getPosition()
        node = self.getEditNode(pos)
        ep = self.getEditNodeIndex(node)

        if ep is None:
            return App.Console.PrintWarning(translate("draft",
                                                      "Node not found")
                                                      + "\n")

        doc = App.getDocument(str(node.documentName.getValue()))
        self.obj = doc.getObject(str(node.objectName.getValue()))
        if self.obj is None:
            return
        if not (Draft.getType(self.obj) in ["Wire", "BSpline", "BezCurve"]):
            return
        if len(self.obj.Points) <= 2:
            App.Console.PrintWarning(translate("draft", 
                                               "Active object must have more than two points/nodes") 
                                               + "\n")
            return

        pts = self.obj.Points
        pts.pop(ep)
        self.obj.Points = pts
        if Draft.getType(self.obj) == "BezCurve":
            self.obj.Proxy.resetcontinuity(self.obj)
        self.obj.recompute()

        # don't do tan/sym on DWire/BSpline!
        self.removeTrackers(self.obj)
        self.setEditPoints(self.obj)

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : GENERAL
    # -------------------------------------------------------------------------

    def setEditPoints(self, obj):
        """append given object's editpoints to self.edipoints and set EditTrackers"""
        self.setPlacement(obj)
        self.editpoints = self.getEditPoints(obj)
        if self.editpoints:  # set trackers and align plane
            self.setTrackers(obj, self.editpoints)
            self.editpoints = []

    def getEditPoints(self, obj):
        """Return a list of App.Vectors relative to object edit nodes.

        (object)
        """
        objectType = Draft.getType(obj)

        if objectType in ["Wire", "BSpline"]:
            self.ui.editUi("Wire")
            return self.getWirePts(obj)
        elif objectType == "BezCurve":
            self.ui.editUi("BezCurve")
            self.resetTrackersBezier(obj)
            self.editpoints = []
            return
        elif objectType == "Circle":
            return self.getCirclePts(obj)
        elif objectType == "Rectangle":
            return self.getRectanglePts(obj)
        elif objectType == "Polygon":
            return self.getPolygonPts(obj)
        elif objectType == "Dimension":
            return self.getDimensionPts(obj)
        elif objectType == "Wall":
            return self.getWallPts(obj)
        elif objectType == "Window":
            return self.getWindowPts(obj)
        elif objectType == "Space":
            return self.getSpacePts(obj)
        elif objectType == "Structure":
            return self.getStructurePts(obj)
        elif objectType == "PanelCut":
            return self.getPanelCutPts(obj)
        elif objectType == "PanelSheet":
            return self.getPanelSheetPts(obj)
        elif objectType == "Part" and obj.TypeId == "Part::Box":
            return self.getPartBoxPts(obj)
        elif objectType == "Part::Line" and obj.TypeId == "Part::Line":
            return self.getPartLinePts(obj)
        elif objectType == "Sketch":
            return self.getSketchPts(obj)
        else:
            return None

    def update(self, obj, nodeIndex, v):
        """Apply the App.Vector to the modified point and update self.obj."""

        objectType = Draft.getType(obj)
        App.ActiveDocument.openTransaction("Edit")

        if objectType in ["Wire", "BSpline"]:
            self.updateWire(obj, nodeIndex, v)
        elif objectType == "BezCurve":
            self.updateWire(obj, nodeIndex, v)
        elif objectType == "Circle":
            self.updateCircle(obj, nodeIndex, v)
        elif objectType == "Rectangle":
            self.updateRectangle(obj, nodeIndex, v)
        elif objectType == "Polygon":
            self.updatePolygon(obj, nodeIndex, v)
        elif objectType == "Dimension":
            self.updateDimension(obj, nodeIndex, v)
        elif objectType == "Sketch":
            self.updateSketch(obj, nodeIndex, v)
        elif objectType == "Wall":
            self.updateWall(obj, nodeIndex, v)
        elif objectType == "Window":
            self.updateWindow(obj, nodeIndex, v)
        elif objectType == "Space":
            self.updateSpace(obj, nodeIndex, v)
        elif objectType == "Structure":
            self.updateStructure(obj, nodeIndex, v)
        elif objectType == "PanelCut":
            self.updatePanelCut(obj, nodeIndex, v)
        elif objectType == "PanelSheet":
            self.updatePanelSheet(obj, nodeIndex, v)
        elif objectType == "Part::Line" and self.obj.TypeId == "Part::Line":
            self.updatePartLine(obj, nodeIndex, v)
        elif objectType == "Part" and self.obj.TypeId == "Part::Box":
            self.updatePartBox(obj, nodeIndex, v)
        obj.recompute()

        App.ActiveDocument.commitTransaction()

        try:
            Gui.ActiveDocument.ActiveView.redraw()
        except AttributeError as err:
            pass

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Line/Wire/Bspline/Bezcurve
    # -------------------------------------------------------------------------

    def getWirePts(self, obj):
        editpoints = []
        for p in obj.Points:
            p = obj.getGlobalPlacement().multVec(p)
            editpoints.append(p)
        return editpoints

    def updateWire(self, obj, nodeIndex, v):
        pts = obj.Points
        editPnt = obj.getGlobalPlacement().inverse().multVec(v)
        # DNC: allows to close the curve by placing ends close to each other
        tol = 0.001
        if ( ( nodeIndex == 0 ) and ( (editPnt - pts[-1]).Length < tol) ) or ( 
                nodeIndex == len(pts) - 1 ) and ( (editPnt - pts[0]).Length < tol):
            obj.Closed = True
        # DNC: fix error message if edited point coincides with one of the existing points
        if ( editPnt in pts ) == True: # checks if point enter is equal to other, this could cause a OCC problem
            App.Console.PrintMessage(translate("draft", 
                                               "This object does not support possible "
                                               "coincident points, please try again.")
                                     + "\n")
            if Draft.getType(obj) in ["BezCurve"]:
                self.resetTrackers(obj)
            else:
                self.trackers[obj.Name][nodeIndex].set(obj.getGlobalPlacement().
                    multVec(obj.Points[nodeIndex]))
            return
        if Draft.getType(obj) in ["BezCurve"]:
            pts = self.recomputePointsBezier(obj,pts,nodeIndex,v,obj.Degree,moveTrackers=False)

        if obj.Closed:
            # check that the new point lies on the plane of the wire
            if hasattr(obj.Shape,"normalAt"):
                normal = obj.Shape.normalAt(0,0)
                point_on_plane = obj.Shape.Vertexes[0].Point
                print(v)
                v.projectToPlane(point_on_plane, normal)
                print(v)
                editPnt = obj.getGlobalPlacement().inverse().multVec(v)
        pts[nodeIndex] = editPnt
        obj.Points = pts
        self.trackers[obj.Name][nodeIndex].set(v)

    def recomputePointsBezier(self, obj, pts, idx, v,
                              degree, moveTrackers=True):
        """
        (object, Points as list, nodeIndex as Int, App.Vector of new point, moveTrackers as Bool)
        return the new point list, applying the App.Vector to the given index point
        """
        editPnt = v
        # DNC: allows to close the curve by placing ends close to each other
        tol = 0.001
        if ( ( idx == 0 ) and ( (editPnt - pts[-1]).Length < tol) ) or (
                idx == len(pts) - 1 ) and ( (editPnt - pts[0]).Length < tol):
            obj.Closed = True
        # DNC: fix error message if edited point coincides with one of the existing points
        #if ( editPnt in pts ) == False:
        knot = None
        ispole = idx % degree

        if ispole == 0: #knot
            if degree >= 3:
                if idx >= 1: #move left pole
                    knotidx = idx if idx < len(pts) else 0
                    pts[idx-1] = pts[idx-1] + editPnt - pts[knotidx]
                    if moveTrackers:
                        self.trackers[obj.Name][idx-1].set(pts[idx-1])
                if idx < len(pts)-1: #move right pole
                    pts[idx+1] = pts[idx+1] + editPnt - pts[idx]
                    if moveTrackers:
                        self.trackers[obj.Name][idx+1].set(pts[idx+1])
                if idx == 0 and obj.Closed: # move last pole
                    pts[-1] = pts [-1] + editPnt -pts[idx]
                    if moveTrackers:
                        self.trackers[obj.Name][-1].set(pts[-1])

        elif ispole == 1 and (idx >=2 or obj.Closed): #right pole
            knot = idx -1
            changep = idx - 2  # -1 in case of closed curve

        elif ispole == degree-1 and idx <= len(pts)-3:  # left pole
            knot = idx + 1
            changep = idx + 2

        elif ispole == degree-1 and obj.Closed and idx == len(pts)-1: #last pole
            knot = 0
            changep = 1

        if knot is not None:  # we need to modify the opposite pole
            segment = int(knot / degree) - 1
            cont = obj.Continuity[segment] if len(obj.Continuity) > segment else 0
            if cont == 1: #tangent
                pts[changep] = obj.Proxy.modifytangentpole(pts[knot],
                    editPnt,pts[changep])
                if moveTrackers:
                    self.trackers[obj.Name][changep].set(pts[changep])
            elif cont == 2: #symmetric
                pts[changep] = obj.Proxy.modifysymmetricpole(pts[knot],editPnt)
                if moveTrackers:
                    self.trackers[obj.Name][changep].set(pts[changep])
        pts[idx] = v

        return pts  # returns the list of new points, taking into account knot continuity

    def resetTrackersBezier(self, obj):
        # in future move tracker definition to DraftTrackers
        from pivy import coin
        knotmarkers = (coin.SoMarkerSet.DIAMOND_FILLED_9_9,#sharp
                coin.SoMarkerSet.SQUARE_FILLED_9_9,        #tangent
                coin.SoMarkerSet.HOURGLASS_FILLED_9_9)     #symmetric
        polemarker = coin.SoMarkerSet.CIRCLE_FILLED_9_9    #pole
        self.trackers[obj.Name] = []
        cont = obj.Continuity
        firstknotcont = cont[-1] if (obj.Closed and cont) else 0
        pointswithmarkers = [(obj.Shape.Edges[0].Curve.
                getPole(1),knotmarkers[firstknotcont])]
        for edgeindex, edge in enumerate(obj.Shape.Edges):
            poles = edge.Curve.getPoles()
            pointswithmarkers.extend([(point,polemarker) for point in poles[1:-1]])
            if not obj.Closed or len(obj.Shape.Edges) > edgeindex +1:
                knotmarkeri = cont[edgeindex] if len(cont) > edgeindex else 0
                pointswithmarkers.append((poles[-1],knotmarkers[knotmarkeri]))
        for index, pwm in enumerate(pointswithmarkers):
            p, marker = pwm
            # if self.pl: p = self.pl.multVec(p)
            self.trackers[obj.Name].append(trackers.editTracker(p,obj.Name,
                index,obj.ViewObject.LineColor,marker=marker))

    def smoothBezPoint(self, obj, point, style='Symmetric'):
        "called when changing the continuity of a knot"
        style2cont = {'Sharp':0,'Tangent':1,'Symmetric':2}
        if point is None:
            return
        if not (Draft.getType(obj) == "BezCurve"):
            return
        pts = obj.Points
        deg = obj.Degree
        if deg < 2:
            return
        if point % deg != 0:  # point is a pole
            if deg >=3:  # allow to select poles
                if (point % deg == 1) and (point > 2 or obj.Closed): #right pole
                    knot = point -1
                    keepp = point
                    changep = point -2
                elif point < len(pts) -3 and point % deg == deg -1: #left pole
                    knot = point +1
                    keepp = point
                    changep = point +2
                elif point == len(pts)-1 and obj.Closed: #last pole
                    # if the curve is closed the last pole has the last
                    # index in the points lists
                    knot = 0
                    keepp = point
                    changep = 1
                else:
                    App.Console.PrintWarning(translate("draft",
                                                       "Can't change Knot belonging to pole %d"%point)
                                                       + "\n")
                    return
                if knot:
                    if style == 'Tangent':
                        pts[changep] = obj.Proxy.modifytangentpole(pts[knot],
                            pts[keepp],pts[changep])
                    elif style == 'Symmetric':
                        pts[changep] = obj.Proxy.modifysymmetricpole(pts[knot],
                            pts[keepp])
                    else: #sharp
                        pass #
            else:
                App.Console.PrintWarning(translate("draft",
                                                   "Selection is not a Knot")
                                                   + "\n")
                return
        else: #point is a knot
            if style == 'Sharp':
                if obj.Closed and point == len(pts)-1:
                    knot = 0
                else:
                    knot = point
            elif style == 'Tangent' and point > 0 and point < len(pts)-1:
                prev, next = obj.Proxy.tangentpoles(pts[point], pts[point-1], pts[point+1])
                pts[point-1] = prev
                pts[point+1] = next
                knot = point  # index for continuity
            elif style == 'Symmetric' and point > 0 and point < len(pts)-1:
                prev, next = obj.Proxy.symmetricpoles(pts[point], pts[point-1], pts[point+1])
                pts[point-1] = prev
                pts[point+1] = next
                knot = point  # index for continuity
            elif obj.Closed and (style == 'Symmetric' or style == 'Tangent'):
                if style == 'Tangent':
                    pts[1], pts[-1] = obj.Proxy.tangentpoles(pts[0], pts[1], pts[-1])
                elif style == 'Symmetric':
                    pts[1], pts[-1] = obj.Proxy.symmetricpoles(pts[0], pts[1], pts[-1])
                knot = 0
            else:
                App.Console.PrintWarning(translate("draft",
                                                       "Endpoint of BezCurve can't be smoothed")
                                                       + "\n")
                return
        segment = knot // deg  # segment index
        newcont = obj.Continuity[:]  # don't edit a property inplace !!!
        if not obj.Closed and (len(obj.Continuity) == segment -1 or
            segment == 0) :
            pass # open curve
        elif (len(obj.Continuity) >= segment or obj.Closed and segment == 0 and
                len(obj.Continuity) >1):
            newcont[segment-1] = style2cont.get(style)
        else: #should not happen
            App.Console.PrintWarning('Continuity indexing error:'
                                     + 'point:%d deg:%d len(cont):%d' % (knot,deg,
                                         len(obj.Continuity)))
        obj.Points = pts
        obj.Continuity = newcont
        self.resetTrackers(obj)

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Rectangle
    # -------------------------------------------------------------------------

    def getRectanglePts(self, obj):
        """Return the list of edipoints for the given Draft Rectangle.

        0 : Placement.Base
        1 : Length
        2 : Height
        """
        editpoints = []
        editpoints.append(obj.getGlobalPlacement().Base)
        editpoints.append(obj.getGlobalPlacement().multVec(App.Vector(obj.Length,0,0)))
        editpoints.append(obj.getGlobalPlacement().multVec(App.Vector(0,obj.Height,0)))
        return editpoints

    def updateRectangleTrackers(self, obj):
        self.trackers[obj.Name][0].set(obj.getGlobalPlacement().Base)
        self.trackers[obj.Name][1].set(obj.getGlobalPlacement().multVec(App.Vector(obj.Length,0,0)))
        self.trackers[obj.Name][2].set(obj.getGlobalPlacement().multVec(App.Vector(0,obj.Height,0)))

    def updateRectangle(self, obj, nodeIndex, v):
        import DraftVecUtils
        delta = obj.getGlobalPlacement().inverse().multVec(v)
        if nodeIndex == 0:
            # p = obj.getGlobalPlacement()
            # p.move(delta)
            obj.Placement.move(delta)
        elif self.editing == 1:
            obj.Length = DraftVecUtils.project(delta,App.Vector(1,0,0)).Length
        elif self.editing == 2:
            obj.Height = DraftVecUtils.project(delta,App.Vector(0,1,0)).Length
        self.updateRectangleTrackers(obj)

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Ellipse (# TODO: yet to be implemented)
    # -------------------------------------------------------------------------

    def setEllipsePts(self):
        return

    def updateEllipse(self,v):
        return

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Circle/Arc
    # -------------------------------------------------------------------------

    def getCirclePts(self, obj):
        """Return the list of edipoints for the given Draft Arc or Circle.

        circle:
        0 : Placement.Base or center
        1 : radius

        arc:
        0 : Placement.Base or center
        1 : first endpoint
        2 : second endpoint
        3 : midpoint
        """
        editpoints = []
        editpoints.append(obj.getGlobalPlacement().Base)
        if obj.FirstAngle == obj.LastAngle:
            # obj is a circle
            self.ui.editUi("Circle")
            editpoints.append(obj.getGlobalPlacement().multVec(App.Vector(obj.Radius,0,0)))
        else:
            # obj is an arc
            self.ui.editUi("Arc")
            editpoints.append(self.getArcStart(obj, global_placement=True))#First endpoint
            editpoints.append(self.getArcEnd(obj, global_placement=True))#Second endpoint
            editpoints.append(self.getArcMid(obj, global_placement=True))#Midpoint
        return editpoints

    def updateCircleTrackers(self, obj):
        self.trackers[obj.Name][0].set(obj.getGlobalPlacement().Base)
        self.trackers[obj.Name][1].set(self.getArcStart(obj, global_placement=True))
        if len(self.trackers[obj.Name]) > 2: 
            # object is an arc
            self.trackers[obj.Name][2].set(self.getArcEnd(obj, global_placement=True))
            self.trackers[obj.Name][3].set(self.getArcMid(obj, global_placement=True))

    def updateCircle(self, obj, nodeIndex, v):
        delta = obj.getGlobalPlacement().inverse().multVec(v)
        local_v = obj.Placement.multVec(delta)

        if obj.FirstAngle == obj.LastAngle:
            # object is a circle
            if nodeIndex == 0:
                obj.Placement.Base = local_v
            elif nodeIndex == 1:
                obj.Radius = delta.Length

        else:
            # obj is an arc
            if self.alt_edit_mode == 0:
                # edit arc by 3 points
                import Part
                if nodeIndex == 0:
                    # center point
                    import DraftVecUtils
                    p1 = self.getArcStart(obj)
                    p2 = self.getArcEnd(obj)
                    p0 = DraftVecUtils.project(delta,self.getArcMid(obj))
                    obj.Radius = p1.sub(p0).Length
                    obj.FirstAngle = -math.degrees(DraftVecUtils.angle(p1.sub(p0)))
                    obj.LastAngle = -math.degrees(DraftVecUtils.angle(p2.sub(p0)))
                    obj.Placement.Base = obj.Placement.multVec(p0)
                    self.setPlacement(obj)

                else:
                    if nodeIndex == 1:  # first point
                        p1=v
                        p2=self.getArcMid(obj,global_placement=True)
                        p3=self.getArcEnd(obj,global_placement=True)
                    elif nodeIndex == 3:  # midpoint
                        p1=self.getArcStart(obj,global_placement=True)
                        p2=v
                        p3=self.getArcEnd(obj,global_placement=True)
                    elif nodeIndex == 2:  # second point
                        p1=self.getArcStart(obj,global_placement=True)
                        p2=self.getArcMid(obj,global_placement=True)
                        p3=v
                    arc=Part.ArcOfCircle(p1,p2,p3)
                    obj.Placement.Base = obj.Placement.multVec(obj.getGlobalPlacement().inverse().multVec(arc.Location))
                    self.setPlacement(obj)
                    obj.Radius = arc.Radius
                    delta = self.invpl.multVec(p1)
                    obj.FirstAngle = math.degrees(math.atan2(delta[1],delta[0]))
                    delta = self.invpl.multVec(p3)
                    obj.LastAngle = math.degrees(math.atan2(delta[1],delta[0]))

            elif self.alt_edit_mode == 1:
                # edit arc by center radius FirstAngle LastAngle
                if nodeIndex == 0:
                    obj.Placement.Base = local_v
                    self.setPlacement(obj)
                else:
                    dangle = math.degrees(math.atan2(delta[1],delta[0]))
                    if nodeIndex == 1:
                        obj.FirstAngle = dangle
                    elif nodeIndex == 2:
                        obj.LastAngle = dangle
                    elif nodeIndex == 3:
                        obj.Radius = delta.Length

        obj.recompute()
        self.updateCircleTrackers(obj)


    def getArcStart(self, obj, global_placement=False):#Returns object midpoint
        if Draft.getType(obj) == "Circle":
            return self.pointOnCircle(obj, obj.FirstAngle, global_placement)

    def getArcEnd(self, obj, global_placement=False):#Returns object midpoint
        if Draft.getType(obj) == "Circle":
            return self.pointOnCircle(obj, obj.LastAngle, global_placement)

    def getArcMid(self, obj, global_placement=False):#Returns object midpoint
        if Draft.getType(obj) == "Circle":
            if obj.LastAngle > obj.FirstAngle:
                midAngle = obj.FirstAngle + (obj.LastAngle - obj.FirstAngle) / 2.0
            else:
                midAngle = obj.FirstAngle + (obj.LastAngle - obj.FirstAngle) / 2.0
                midAngle += App.Units.Quantity(180,App.Units.Angle)
            return self.pointOnCircle(obj, midAngle, global_placement)

    def pointOnCircle(self, obj, angle, global_placement=False):
        if Draft.getType(obj) == "Circle":
            px = obj.Radius * math.cos(math.radians(angle))
            py = obj.Radius * math.sin(math.radians(angle))
            p = App.Vector(px, py, 0.0)
            if global_placement == True:
                p = obj.getGlobalPlacement().multVec(p)
            return p
        return None

    def arcInvert(self, obj):
        obj.FirstAngle, obj.LastAngle = obj.LastAngle, obj.FirstAngle
        obj.recompute()
        self.updateCircleTrackers(obj)

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Polygon (maybe could also rotate the polygon)
    # -------------------------------------------------------------------------

    def getPolygonPts(self, obj):
        editpoints = []
        editpoints.append(obj.Placement.Base)
        editpoints.append(obj.Shape.Vertexes[0].Point)
        return editpoints

    def updatePolygon(self, obj, nodeIndex, v):
        delta = v.sub(self.obj.Placement.Base)
        if self.editing == 0:
            p = self.obj.Placement
            p.move(delta)
            self.obj.Placement = p
            self.trackers[self.obj.Name][0].set(self.obj.Placement.Base)
        elif self.editing == 1:
            if self.obj.DrawMode == 'inscribed':
                self.obj.Radius = delta.Length
            else:
                halfangle = ((math.pi*2)/self.obj.FacesNumber)/2
                rad = math.cos(halfangle)*delta.Length
                self.obj.Radius = rad
            self.obj.recompute()
        self.trackers[self.obj.Name][1].set(self.obj.Shape.Vertexes[0].Point)

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : Dimension (point on dimension line is not clickable)
    # -------------------------------------------------------------------------

    def getDimensionPts(self, obj):
        editpoints = []
        p = obj.ViewObject.Proxy.textpos.translation.getValue()
        editpoints.append(obj.Start)
        editpoints.append(obj.End)
        editpoints.append(obj.Dimline)
        editpoints.append(App.Vector(p[0], p[1], p[2]))
        return editpoints

    def updateDimension(self, obj, nodeIndex, v):
        if self.editing == 0:
            self.obj.Start = v
        elif self.editing == 1:
            self.obj.End = v
        elif self.editing == 2:
            self.obj.Dimline = v
        elif self.editing == 3:
            self.obj.ViewObject.TextPosition = v

    # -------------------------------------------------------------------------
    # EDIT OBJECT TOOLS : ARCH Wall, Windows, Structure, Panel, etc.
    # -------------------------------------------------------------------------

    # SKETCH: just if it's composed by a single segment-----------------------

    def getSketchPts(self, obj):
        """Return the list of edipoints for the given single line sketch.

        (WallTrace)
        0 : startpoint
        1 : endpoint
        """
        editpoints = []
        if obj.GeometryCount == 1:
            editpoints.append(obj.getGlobalPlacement().multVec(obj.getPoint(0,1)))
            editpoints.append(obj.getGlobalPlacement().multVec(obj.getPoint(0,2)))
            return editpoints
        else:
            App.Console.PrintWarning(translate("draft",
                                               "Sketch is too complex to edit: "
                                               "it is suggested to use sketcher default editor")
                                     + "\n")
            return None

    def updateSketch(self, obj, nodeIndex, v):
        """Move a single line sketch vertex a certain displacement.

        (single segment sketch object, node index as Int, App.Vector)
        move a single line sketch (WallTrace) vertex according to a given App.Vector
        0 : startpoint
        1 : endpoint
        """
        if nodeIndex == 0:
            obj.movePoint(0,1,obj.getGlobalPlacement().inverse().multVec(v))
        elif nodeIndex == 1:
            obj.movePoint(0,2,obj.getGlobalPlacement().inverse().multVec(v))
        obj.recompute()

    # WALL---------------------------------------------------------------------

    def getWallPts(self, obj):
        """Return the list of edipoints for the given Arch Wall object.

        0 : height of the wall
        1-to end : base object editpoints, in place with the wall
        """
        editpoints = []
        # height of the wall
        editpoints.append(obj.getGlobalPlacement().multVec(App.Vector(0,0,obj.Height)))
        # try to add here an editpoint based on wall height (maybe should be good to associate it with a circular tracker)
        if obj.Base:
            # base points are added to self.trackers under wall-name key
            basepoints = []
            if Draft.getType(obj.Base) in ["Wire","Circle","Rectangle",
                                            "Polygon", "Sketch"]:
                basepoints = self.getEditPoints(obj.Base)
                for point in basepoints:
                    editpoints.append(obj.Placement.multVec(point)) #works ok except if App::Part is rotated... why?
        return editpoints

    def updateWallTrackers(self, obj):
        """Update self.trackers[obj.Name][0] to match with given object."""
        pass

    def updateWall(self, obj, nodeIndex, v):
        import DraftVecUtils
        if nodeIndex == 0:
            delta= obj.getGlobalPlacement().inverse().multVec(v)
            vz=DraftVecUtils.project(delta,App.Vector(0, 0, 1))
            if vz.Length > 0:
                obj.Height = vz.Length
        elif nodeIndex > 0:
            if obj.Base:
                if Draft.getType(obj.Base) in ["Wire", "Circle", "Rectangle",
                                               "Polygon", "Sketch"]:
                    self.update(obj.Base, nodeIndex - 1, 
                        obj.Placement.inverse().multVec(v))
        obj.recompute()

    # WINDOW-------------------------------------------------------------------

    def getWindowPts(self, obj):
        import DraftGeomUtils
        editpoints = []
        pos = obj.Base.Placement.Base
        h = float(obj.Height) + pos.z
        normal = obj.Normal
        angle = normal.getAngle(App.Vector(1, 0, 0))
        editpoints.append(pos)
        editpoints.append(App.Vector(pos.x + float(obj.Width) * math.cos(angle-math.pi / 2.0),
                                                pos.y + float(obj.Width) * math.sin(angle-math.pi / 2.0),
                                                pos.z))
        editpoints.append(App.Vector(pos.x, pos.y, h))
        return editpoints

    def updateWindow(self, obj, nodeIndex, v):
        pos = self.obj.Base.Placement.Base
        if self.editing == 0:
            self.obj.Base.Placement.Base = v
            self.obj.Base.recompute()
        if self.editing == 1:
            self.obj.Width = pos.sub(v).Length
            self.obj.Base.recompute()
        if self.editing == 2:
            self.obj.Height = pos.sub(v).Length
            self.obj.Base.recompute()
        for obj in self.obj.Hosts:
            obj.recompute()
        self.obj.recompute()

    # STRUCTURE----------------------------------------------------------------

    def getStructurePts(self, obj):
        if obj.Nodes:
            editpoints = []
            self.originalDisplayMode = obj.ViewObject.DisplayMode
            self.originalPoints = obj.ViewObject.NodeSize
            self.originalNodes = obj.ViewObject.ShowNodes
            self.obj.ViewObject.DisplayMode = "Wireframe"
            self.obj.ViewObject.NodeSize = 1
            # self.obj.ViewObject.ShowNodes = True
            for p in obj.Nodes:
                if self.pl:
                    p = self.pl.multVec(p)
                editpoints.append(p)
            return editpoints
        else:
            return None

    def updateStructure(self, obj, nodeIndex, v):
        nodes = self.obj.Nodes
        nodes[self.editing] = self.invpl.multVec(v)
        self.obj.Nodes = nodes

    # SPACE--------------------------------------------------------------------

    def getSpacePts(self, obj):
        try:
            editpoints = []
            self.editpoints.append(obj.ViewObject.Proxy.getTextPosition(obj.ViewObject))
            return editpoints
        except:
            pass

    def updateSpace(self, obj, nodeIndex, v):
        if self.editing == 0:
            self.obj.ViewObject.TextPosition = v

    # PANELS-------------------------------------------------------------------

    def getPanelCutPts(self, obj):
        editpoints = []
        if self.obj.TagPosition.Length == 0:
            pos = obj.Shape.BoundBox.Center
        else:
            pos = self.pl.multVec(obj.TagPosition)
        editpoints.append(pos)
        return editpoints

    def updatePanelCut(self, obj, nodeIndex, v):
        if self.editing == 0:
            self.obj.TagPosition = self.invpl.multVec(v)

    def getPanelSheetPts(self, obj):
        editpoints = []
        editpoints.append(self.pl.multVec(obj.TagPosition))
        for o in obj.Group:
            editpoints.append(self.pl.multVec(o.Placement.Base))
        return editpoints

    def updatePanelSheet(self, obj, nodeIndex, v):
        if self.editing == 0:
            self.obj.TagPosition = self.invpl.multVec(v)
        else:
            self.obj.Group[self.editing-1].Placement.Base = self.invpl.multVec(v)
    
    # PART::LINE--------------------------------------------------------------

    def getPartLinePts(self, obj):
        editpoints = []
        editpoints.append(self.pl.multVec(App.Vector(obj.X1,obj.Y1,obj.Z1)))
        editpoints.append(self.pl.multVec(App.Vector(obj.X2,obj.Y2,obj.Z2)))
        return editpoints
    
    def updatePartLine(self, obj, nodeIndex, v):
        pt=self.invpl.multVec(v)
        if self.editing == 0:
            self.obj.X1 = pt.x
            self.obj.Y1 = pt.y
            self.obj.Z1 = pt.z
        elif self.editing == 1:
            self.obj.X2 = pt.x
            self.obj.Y2 = pt.y
            self.obj.Z2 = pt.z

    # PART::BOX---------------------------------------------------------------

    def getPartBoxPts(self, obj):
        editpoints = []
        editpoints.append(obj.Placement.Base)
        editpoints.append(self.pl.multVec(App.Vector(obj.Length, 0, 0)))
        editpoints.append(self.pl.multVec(App.Vector(0, obj.Width, 0)))
        editpoints.append(self.pl.multVec(App.Vector(0, 0, obj.Height)))
        return editpoints

    def updatePartBox(self, obj, nodeIndex, v):
        import DraftVecUtils
        delta = self.invpl.multVec(v)
        if self.editing == 0:
            self.obj.Placement.Base = v
            self.setPlacement(self.obj)
        elif self.editing == 1:
            xApp.Vector = DraftVecUtils.project(delta, App.Vector(1, 0, 0))
            self.obj.Length = xApp.Vector.Length
        elif self.editing == 2:
            xApp.Vector = DraftVecUtils.project(delta, App.Vector(0, 1, 0))
            self.obj.Width = xApp.Vector.Length
        elif self.editing == 3:
            xApp.Vector = DraftVecUtils.project(delta, App.Vector(0, 0, 1))
            self.obj.Height = xApp.Vector.Length
        self.trackers[self.obj.Name][0].set(self.obj.Placement.Base)
        self.trackers[self.obj.Name][1].set(self.pl.multVec(App.Vector(self.obj.Length,0,0)))
        self.trackers[self.obj.Name][2].set(self.pl.multVec(App.Vector(0,self.obj.Width,0)))
        self.trackers[self.obj.Name][3].set(self.pl.multVec(App.Vector(0,0,self.obj.Height)))

    # ------------------------------------------------------------------------
    # Context menu
    # ------------------------------------------------------------------------

    def display_tracker_menu(self, event):
        self.tracker_menu = QtGui.QMenu()
        self.event = event
        actions = None
        if self.overNode:
            # if user is over a node
            doc = self.overNode.get_doc_name()
            obj = App.getDocument(doc).getObject(self.overNode.get_obj_name())
            ep = self.overNode.get_subelement_index()
            if Draft.getType(obj) in ["Line", "Wire"]:
                actions = ["delete point"]
            elif Draft.getType(obj) in ["Circle"]:
                if obj.FirstAngle != obj.LastAngle:
                    if ep == 0:  # user is over arc start point
                        actions = ["move arc"]
                    elif ep == 1:  # user is over arc start point
                        actions = ["set first angle"]
                    elif ep == 2:  # user is over arc end point
                        actions = ["set last angle"]
                    elif ep == 3:  # user is over arc mid point
                        actions = ["set radius"]
            elif Draft.getType(obj) in ["BezCurve"]:
                actions = ["make sharp", "make tangent",
                           "make symmetric", "delete point"]
            else:
                return
        else:
            # if user is over an edited object
            pos = self.event.getPosition()
            obj = self.get_selected_obj_at_position(pos)
            if Draft.getType(obj) in ["Line", "Wire", "BSpline", "BezCurve"]:
                actions = ["add point"]
            elif Draft.getType(obj) in ["Circle"] and obj.FirstAngle != obj.LastAngle:
                actions = ["invert arc"]
        if actions is None:
            return
        for a in actions:
            self.tracker_menu.addAction(a)
        self.tracker_menu.popup(Gui.getMainWindow().cursor().pos())
        QtCore.QObject.connect(self.tracker_menu,QtCore.SIGNAL("triggered(QAction *)"),self.evaluate_menu_action)

    def evaluate_menu_action(self, labelname):
        action_label = str(labelname.text())
        # Bezier curve menu
        if action_label in ["make sharp", "make tangent", "make symmetric"]:
            doc = self.overNode.get_doc_name()
            obj = App.getDocument(doc).getObject(self.overNode.get_obj_name())
            idx = self.overNode.get_subelement_index()
            if action_label == "make sharp":
                self.smoothBezPoint(obj, idx, 'Sharp')
            elif action_label == "make tangent":
                self.smoothBezPoint(obj, idx, 'Tangent')
            elif action_label == "make symmetric":
                self.smoothBezPoint(obj, idx, 'Symmetric')
        # addPoint and deletePoint menu
        elif action_label == "delete point":
            self.delPoint(self.event)
        elif action_label == "add point":
            self.addPoint(self.event)
        # arc tools
        elif action_label in ("move arc", "set radius",
                              "set first angle", "set last angle"):
            self.alt_edit_mode = 1
            self.startEditing(self.event)
        elif action_label == "invert arc":
            pos = self.event.getPosition()
            obj = self.get_selected_obj_at_position(pos)
            self.arcInvert(obj)
        del self.event


Gui.addCommand('Draft_Edit', Edit())



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function

import FreeCAD
import FreeCADGui
import Path
import PathScripts
import PathScripts.PathLog as PathLog
import PathScripts.PathUtil as PathUtil
import json
import os
import xml.sax

from PySide import QtCore, QtGui

PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())
#PathLog.trackModule(PathLog.thisModule())

def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)

# Tooltable XML readers
class FreeCADTooltableHandler(xml.sax.ContentHandler):
    # http://www.tutorialspoint.com/python/python_xml_processing.htm

    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.tooltable = None
        self.tool = None
        self.number = None

    # Call when an element is found
    def startElement(self, name, attrs):
        if name == "Tooltable":
            self.tooltable = Path.Tooltable()
        elif name == "Toolslot":
            self.number = int(attrs["number"])
        elif name == "Tool":
            self.tool = Path.Tool()
            self.tool.Name = str(attrs["name"])
            self.tool.ToolType = str(attrs["type"])
            self.tool.Material = str(attrs["mat"])
            # for some reason without the following line I get an error
            #print attrs["diameter"]
            self.tool.Diameter = float(attrs["diameter"])
            self.tool.LengthOffset = float(attrs["length"])
            self.tool.FlatRadius = float(attrs["flat"])
            self.tool.CornerRadius = float(attrs["corner"])
            self.tool.CuttingEdgeAngle = float(attrs["angle"])
            self.tool.CuttingEdgeHeight = float(attrs["height"])

    # Call when an elements ends
    def endElement(self, name):
        if name == "Toolslot":
            if self.tooltable and self.tool and self.number:
                self.tooltable.setTool(self.number, self.tool)
                self.number = None
                self.tool = None


class HeeksTooltableHandler(xml.sax.ContentHandler):

    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.tooltable = Path.Tooltable()
        self.tool = None
        self.number = None

    # Call when an element is found
    def startElement(self, name, attrs):
        if name == "Tool":
            self.tool = Path.Tool()
            self.number = int(attrs["tool_number"])
            self.tool.Name = str(attrs["title"])
        elif name == "params":
            t = str(attrs["type"])
            if t == "drill":
                self.tool.ToolType = "Drill"
            elif t == "center_drill_bit":
                self.tool.ToolType = "CenterDrill"
            elif t == "end_mill":
                self.tool.ToolType = "EndMill"
            elif t == "slot_cutter":
                self.tool.ToolType = "SlotCutter"
            elif t == "ball_end_mill":
                self.tool.ToolType = "BallEndMill"
            elif t == "chamfer":
                self.tool.ToolType = "Chamfer"
            elif t == "engraving_bit":
                self.tool.ToolType = "Engraver"
            m = str(attrs["material"])
            if m == "0":
                self.tool.Material = "HighSpeedSteel"
            elif m == "1":
                self.tool.Material = "Carbide"
            # for some reason without the following line I get an error
            #print attrs["diameter"]
            self.tool.Diameter = float(attrs["diameter"])
            self.tool.LengthOffset = float(attrs["tool_length_offset"])
            self.tool.FlatRadius = float(attrs["flat_radius"])
            self.tool.CornerRadius = float(attrs["corner_radius"])
            self.tool.CuttingEdgeAngle = float(
                attrs["cutting_edge_angle"])
            self.tool.CuttingEdgeHeight = float(
                attrs["cutting_edge_height"])

    # Call when an elements ends
    def endElement(self, name):
        if name == "Tool":
            if self.tooltable and self.tool and self.number:
                self.tooltable.setTool(self.number, self.tool)
                self.number = None
                self.tool = None


class ToolLibraryManager():
    '''
    The Tool Library is a list of individual tool tables.  Each
    Tool Table can contain n tools.  The tool library will be persisted to user
    preferences and all or part of the library can be exported to other formats
    '''

    TooltableTypeJSON     = translate("PathToolLibraryManager", "Tooltable JSON (*.json)")
    TooltableTypeXML      = translate("PathToolLibraryManager", "Tooltable XML (*.xml)")
    TooltableTypeHeekscad = translate("PathToolLibraryManager", "HeeksCAD tooltable (*.tooltable)")
    TooltableTypeLinuxCNC = translate("PathToolLibraryManager", "LinuxCNC tooltable (*.tbl)")

    PreferenceMainLibraryXML = "ToolLibrary"
    PreferenceMainLibraryJSON = "ToolLibrary-Main"

    def __init__(self):
        self.prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Path")
        self.toolTables = []
        self.currentTableName = None
        self.loadToolTables()

    def getToolTables(self):
        ''' Return tool table list '''
        return self.toolTables

    def getCurrentTableName(self):
        ''' return the name of the currently loaded tool table '''
        return self.currentTableName

    def getCurrentTable(self):
        ''' returns an object of the current tool table '''
        return self.getTableFromName(self.currentTableName)
    
    def getTableFromName(self, name):
        ''' get the tool table object from the name '''
        for table in self.toolTables:
            if table.Name == name:
                return table

    def getNextToolTableName(self, tableName='Tool Table'):
        ''' get a unique name for a new tool table '''
        iter = 1
        tempName = tableName[-2:] 

        if tempName[0] == '-' and tempName[-1].isdigit():
            tableName = tableName[:-2]

        while any(table.Name == tableName + '-' + str(iter) for table in self.toolTables):
            iter += 1              
        
        return tableName + '-' + str(iter) 

    def addNewToolTable(self):
        ''' creates a new tool table '''
        tt = Path.Tooltable()
        tt.Version = 1
        name = self.getNextToolTableName()
        tt.Name = name
        self.toolTables.append(tt)
        self.saveMainLibrary()
        return name

    def deleteToolTable(self):
        ''' deletes the selected tool table '''
        if len(self.toolTables):
            index = next((index for (index, d) in enumerate(self.toolTables) if d.Name == self.currentTableName), None) 
            self.toolTables.pop(index)
            self.saveMainLibrary()

    def renameToolTable(self, newName, index):
        ''' renames a tool table with the new name'''
        currentTableName = self.toolTables[index].Name
        if newName == currentTableName:
            PathLog.error(translate('PathToolLibraryManager', "Tool Table Same Name"))
            return False
        if newName in self.toolTables:
            PathLog.error(translate('PathToolLibraryManager', "Tool Table Name Exists"))
            return False
        tt = self.getTableFromName(currentTableName)
        if tt:
            tt.Name = newName
            self.saveMainLibrary()
            return True
        

    def templateAttrs(self):
        ''' gets the tool table arributes '''
        toolTables = []
        for tt in self.toolTables:
            tableData = {}
            tableData['Version'] = 1
            tableData['TableName'] = tt.Name

            toolData = {}
            for tool in tt.Tools:
                toolData[tool] = tt.Tools[tool].templateAttrs()

            tableData['Tools'] = toolData
            toolTables.append(tableData)
                  
        return toolTables

    def tooltableFromAttrs(self, stringattrs):
        if stringattrs.get('Version') and 1 == int(stringattrs['Version']):
            
            tt = Path.Tooltable()
            tt.Version = 1
            tt.Name = self.getNextToolTableName()

            if stringattrs.get('Version'):
                tt.Version = stringattrs.get('Version')

            if stringattrs.get('TableName'):
                tt.Name = stringattrs.get('TableName')
                if any(table.Name == tt.Name for table in self.toolTables):
                    tt.Name = self.getNextToolTableName(tt.Name)
            
            for key, attrs in PathUtil.keyValueIter(stringattrs['Tools']):
                    tool = Path.Tool() 
                    tool.Name = str(attrs["name"])
                    tool.ToolType = str(attrs["tooltype"])
                    tool.Material = str(attrs["material"])
                    tool.Diameter = float(attrs["diameter"])
                    tool.LengthOffset = float(attrs["lengthOffset"])
                    tool.FlatRadius = float(attrs["flatRadius"])
                    tool.CornerRadius = float(attrs["cornerRadius"])
                    tool.CuttingEdgeAngle = float(attrs["cuttingEdgeAngle"])
                    tool.CuttingEdgeHeight = float(attrs["cuttingEdgeHeight"])
                    tt.setTool(int(key), tool)

            return tt
        else:
            PathLog.error(translate('PathToolLibraryManager', "Unsupported Path tooltable template version %s") % stringattrs.get('Version'))
        return None

    def loadToolTables(self):
        ''' loads the tool tables from the stored data '''
        self.toolTables = []
        self.currentTableName = ''

        def addTable(tt):
            if tt:
                self.toolTables.append(tt)
            else:
                PathLog.error(translate('PathToolLibraryManager', "Unsupported Path tooltable"))

        prefString = self.prefs.GetString(self.PreferenceMainLibraryJSON, "")

        if not prefString:
            return

        prefsData = json.loads(prefString)

        if isinstance(prefsData, dict):
            tt = self.tooltableFromAttrs(prefsData)
            addTable(tt)

        if isinstance(prefsData, list):
            for table in prefsData:
                tt = self.tooltableFromAttrs(table)
                addTable(tt)

        if len(self.toolTables):
            self.currentTableName = self.toolTables[0].Name

    def saveMainLibrary(self):
        '''Persists the permanent library to FreeCAD user preferences'''
        tmpstring = json.dumps(self.templateAttrs())
        self.prefs.SetString(self.PreferenceMainLibraryJSON, tmpstring)
        self.loadToolTables()
        return True

    def getJobList(self):
        '''Builds the list of all Tool Table lists'''
        tablelist = []

        for o in FreeCAD.ActiveDocument.Objects:
            if hasattr(o, "Proxy"):
                if isinstance(o.Proxy, PathScripts.PathJob.ObjectJob):
                    tablelist.append(o.Label)
                    
        return tablelist

    def getTool(self, listname, toolnum):
        ''' gets the tool object '''
        tt = self.getTableFromName(listname)
        return tt.getTool(toolnum)

    def getTools(self, tablename):
        '''returns the tool data for a given table'''
        tooldata = []
        tableExists =  any(table.Name == tablename for table in self.toolTables)
        if tableExists:
            self.currentTableName = tablename
        else:
            return None

        tt = self.getTableFromName(tablename)
        headers = ["","Tool Num.","Name","Tool Type","Diameter"]
        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(headers)

        def unitconv(ivalue):
            val = FreeCAD.Units.Quantity(ivalue, FreeCAD.Units.Length)
            displayed_val = val.UserString      #just the displayed value-not the internal one
            return displayed_val

        if tt:
            if len(tt.Tools) == 0:
                tooldata.append([])
            for number, t in PathUtil.keyValueIter(tt.Tools):

                itemcheck = QtGui.QStandardItem()
                itemcheck.setCheckable(True)
                itemNumber =  QtGui.QStandardItem(str(number))
                itemName =  QtGui.QStandardItem(t.Name)
                itemToolType =  QtGui.QStandardItem(t.ToolType)
                itemDiameter =  QtGui.QStandardItem(unitconv(t.Diameter))

                row = [itemcheck, itemNumber, itemName, itemToolType, itemDiameter]
                model.appendRow(row)

        return model

    # methods for importing and exporting
    def read(self, filename, listname):
        "imports a tooltable from a file"

        importedTables = []

        try:
            fileExtension = os.path.splitext(filename[0])[1].lower()
            xmlHandler = None
            if fileExtension == '.tooltable':
                xmlHandler = HeeksTooltableHandler()
            if fileExtension == '.xml':
                xmlHandler = FreeCADTooltableHandler()

            if xmlHandler:
                parser = xml.sax.make_parser()
                parser.setFeature(xml.sax.handler.feature_namespaces, 0)
                parser.setContentHandler(xmlHandler)
                parser.parse(PathUtil.toUnicode(filename[0]))
                if not xmlHandler.tooltable:
                    return None

                ht = xmlHandler.tooltable
            else:
                with open(PathUtil.toUnicode(filename[0]), "rb") as fp:
                    tableData = json.load(fp)

                    if isinstance(tableData, dict):
                        ht = self.tooltableFromAttrs(tableData)
                        if ht:
                            importedTables.append(ht)

                    if isinstance(tableData, list):
                        for table in tableData:
                            ht = self.tooltableFromAttrs(table)                 
                            if ht:
                                importedTables.append(ht)

            if importedTables:
                for tt in importedTables:
                    self.toolTables.append(tt)
                
                self.saveMainLibrary()
                return True
            else:
                return False

        except Exception as e: # pylint: disable=broad-except
            print("could not parse file", e)


    def write(self, filename, listname):
        "exports the tooltable to a file"
        tt = self.getTableFromName(listname)
        if tt:
            try:
                def openFileWithExtension(name, ext):
                    fext = os.path.splitext(name)[1].lower()
                    if fext != ext:
                        name = "{}{}".format(name, ext)
                    return (open(PathUtil.toUnicode(name), 'w'), name)

                if filename[1] == self.TooltableTypeXML:
                    fp,fname = openFileWithExtension(filename[0], '.xml')
                    fp.write('<?xml version="1.0" encoding="UTF-8"?>\n')
                    fp.write(tt.Content)
                elif filename[1] == self.TooltableTypeLinuxCNC:
                    fp,fname = openFileWithExtension(filename[0], '.tbl')
                    for key in tt.Tools:
                        t = tt.Tools[key]
                        fp.write("T{0} P{0} Y{1} Z{2} A{3} B{4} C{5} U{6} V{7} W{8} D{9} I{10} J{11} Q{12} ;{13}\n".format(key,0,t.LengthOffset,0,0,0,0,0,0,t.Diameter,0,0,0,t.Name))
                else:
                    fp,fname = openFileWithExtension(filename[0], '.json')
                    json.dump(self.templateAttrs(), fp, sort_keys=True, indent=2)

                fp.close()
                print("Written ", PathUtil.toUnicode(fname))

            except Exception as e: # pylint: disable=broad-except
                print("Could not write file:", e)

    def addnew(self, listname, tool, position = None):
        "adds a new tool at the end of the table"
        tt = self.getTableFromName(listname)
        if not tt:
            tt = Path.Tooltable()
        if position is None:
            tt.addTools(tool)
            newID = list(tt.Tools)[-1]
        else:
            tt.setTool(position, tool)
            newID = position

        if listname == self.getCurrentTableName():
            self.saveMainLibrary()
        return newID

    def updateTool(self, listname, toolnum, tool):
        '''updates tool data'''
        tt = self.getTableFromName(listname)
        tt.deleteTool(toolnum)
        tt.setTool(toolnum, tool)
        if listname == self.getCurrentTableName():
            return self.saveMainLibrary()
        return True

    def moveup(self, number, listname):
        "moves a tool to a lower number, if possible"
        target = number - 1
        if number < 2:
            return False, target
        target = number - 1
        tt = self.getTableFromName(listname)
        t1 = tt.getTool(number).copy()
        tt.deleteTool(number)
        if target in tt.Tools.keys():
            t2 = tt.getTool(target).copy()
            tt.deleteTool(target)
            tt.setTool(number, t2)
        tt.setTool(target, t1)
        if listname == self.getCurrentTableName():
            self.saveMainLibrary()
        return True, target

    def movedown(self, number, listname):
        "moves a tool to a higher number, if possible"
        tt = self.getTableFromName(listname)
        target = number + 1
        t1 = tt.getTool(number).copy()
        tt.deleteTool(number)
        if target in tt.Tools.keys():
            t2 = tt.getTool(target).copy()
            tt.deleteTool(target)
            tt.setTool(number, t2)
        tt.setTool(target, t1)
        if listname == self.getCurrentTableName():
            self.saveMainLibrary()
        return True, target

    def duplicate(self, number, listname):
        ''' duplicates the selected tool in the selected tool table '''
        tt = self.getTableFromName(listname)
        tool = tt.getTool(number).copy()
        tt.addTools(tool)

        newID = list(tt.Tools)[-1]

        if listname == self.getCurrentTableName():
            self.saveMainLibrary()
        return True, newID

    def moveToTable(self, number, listname):
        ''' Moves the tool to selected tool table '''
        fromTable = self.getTableFromName(self.getCurrentTableName())
        toTable = self.getTableFromName(listname)
        tool = fromTable.getTool(number).copy()
        toTable.addTools(tool)
        fromTable.deleteTool(number)

    def delete(self, number, listname):
        '''deletes a tool from the current list'''
        tt = self.getTableFromName(listname)
        tt.deleteTool(number)
        if listname == self.getCurrentTableName():
            self.saveMainLibrary()
        return True


# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""This module provides the code for Draft make_ellipse function.
"""
## @package make_ellipse
# \ingroup DRAFT
# \brief This module provides the code for Draft make_ellipse function.

import FreeCAD as App

from draftutils.gui_utils import format_object
from draftutils.gui_utils import select

from draftobjects.ellipse import Ellipse
if App.GuiUp:
    from draftviewproviders.view_base import ViewProviderDraft


def make_ellipse(majradius, minradius, placement=None, face=True, support=None):
    """make_ellipse(majradius, minradius, [placement], [face], [support])
    
    Makes an ellipse with the given major and minor radius, and optionally
    a placement.

    Parameters
    ----------
    majradius : 
        Major radius of the ellipse.

    minradius : 
        Minor radius of the ellipse.

    placement : Base.Placement
        If a placement is given, it is used.
    
    face : Bool
        If face is False, the rectangle is shown as a wireframe, 
        otherwise as a face.   

    support : 
        TODO: Describe.
    """
    if not App.ActiveDocument:
        App.Console.PrintError("No active document. Aborting\n")
        return
    obj = App.ActiveDocument.addObject("Part::Part2DObjectPython", "Ellipse")
    Ellipse(obj)
    if minradius > majradius:
        majradius, minradius = minradius, majradius
    obj.MajorRadius = majradius
    obj.MinorRadius = minradius
    obj.Support = support
    if placement:
        obj.Placement = placement
    if App.GuiUp:
        ViewProviderDraft(obj.ViewObject)
        #if not face:
        #    obj.ViewObject.DisplayMode = "Wireframe"
        format_object(obj)
        select(obj)

    return obj


makeEllipse = make_ellipse


# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provide the Draft_Snap commands used by the snapping mechanism in Draft."""
## @package gui_snaps
# \ingroup DRAFT
# \brief Provide the Draft_Snap commands used by the snapping mechanism
# in Draft.

from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui

import draftguitools.gui_base as gui_base
from draftutils.translate import _tr


# UTILITIES -----------------------------------------------------------------


def get_snap_statusbar_widget():
    """Return snap statusbar button."""
    mw = Gui.getMainWindow()
    if mw:
        sb = mw.statusBar()
        if sb:
            return sb.findChild(QtGui.QToolBar,"draft_snap_widget")
    return None


def sync_snap_toolbar_button(button, status):
    """Set snap toolbar button to given state."""
    snap_toolbar = Gui.Snapper.get_snap_toolbar()
    if not snap_toolbar:
        return
    for a in snap_toolbar.actions():
        if a.objectName() == button:
            if button == "Draft_Snap_Lock_Button":
                # for lock button
                snap_toolbar.actions()[0].setChecked(status)
                for a in snap_toolbar.actions()[1:]:
                    if a.objectName()[:10] == "Draft_Snap":
                        a.setEnabled(status)
            else:
                # for every other button
                a.setChecked(status)
                if a.isChecked():
                    a.setToolTip(a.toolTip().replace("OFF","ON"))
                else:
                    a.setToolTip(a.toolTip().replace("ON","OFF"))


def sync_snap_statusbar_button(button, status):
    """Set snap statusbar button to given state."""
    ssw = get_snap_statusbar_widget()
    if not ssw:
        return
    for child in ssw.children():
        if child.objectName() == "Snap_Statusbutton":
            ssb = child
    actions = []
    for a in ssb.menu().actions() + ssw.children()[-6:]:
        actions.append(a)

    if button == "Draft_Snap_Lock_Statusbutton":
        ssb.setChecked(status)
        for a in actions[1:]:
            if a.objectName()[:10] == "Draft_Snap":
                a.setEnabled(status)
    else:
        for a in actions:
            if a.objectName() == button:
                a.setChecked(status)


# SNAP GUI TOOLS ------------------------------------------------------------


class Draft_Snap_Lock(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Lock tool.

    Activate or deactivate all snap methods at once.
    """

    def __init__(self):
        super(Draft_Snap_Lock, self).__init__(name=_tr("Main toggle snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Main snapping toggle On/Off"
        _tip = ("Activates or deactivates "
                "all snap methods at once.")

        return {'Pixmap': 'Snap_Lock',
                'Accel': "Shift+S",
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Lock", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Lock", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Lock, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Lock')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Lock"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Lock"+"_Statusbutton", status)


Gui.addCommand('Draft_Snap_Lock', Draft_Snap_Lock())


class Draft_Snap_Midpoint(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Midpoint tool.

    Set snapping to the midpoint of an edge.
    """

    def __init__(self):
        super(Draft_Snap_Midpoint, self).__init__(name=_tr("Midpoint snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Midpoint"
        _tip = "Set snapping to the midpoint of an edge."

        return {'Pixmap': 'Snap_Midpoint',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Midpoint", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Midpoint", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Midpoint, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Midpoint')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Midpoint"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Midpoint_Statusbutton", status)


Gui.addCommand('Draft_Snap_Midpoint', Draft_Snap_Midpoint())


class Draft_Snap_Perpendicular(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Perpendicular tool.

    Set snapping to a direction that is perpendicular to an edge.
    """

    def __init__(self):
        super(Draft_Snap_Perpendicular, self).__init__(name=_tr("Perpendicular snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Perpendicular"
        _tip = "Set snapping to a direction that is perpendicular to an edge."

        return {'Pixmap': 'Snap_Perpendicular',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Perpendicular",
                                              _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Perpendicular",
                                             _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Perpendicular, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Perpendicular')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Perpendicular"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Perpendicular_Statusbutton", status)


Gui.addCommand('Draft_Snap_Perpendicular', Draft_Snap_Perpendicular())


class Draft_Snap_Grid(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Grid tool.

    Set snapping to the intersection of grid lines.
    """

    def __init__(self):
        super(Draft_Snap_Grid, self).__init__(name=_tr("Grid snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = "Set snapping to the intersection of grid lines."

        return {'Pixmap': 'Snap_Grid',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Grid", "Grid"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Grid", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Grid, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Grid')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Grid"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Grid_Statusbutton", status)


Gui.addCommand('Draft_Snap_Grid', Draft_Snap_Grid())


class Draft_Snap_Intersection(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Intersection tool.

    Set snapping to the intersection of edges.
    """

    def __init__(self):
        super(Draft_Snap_Intersection, self).__init__(name=_tr("Intersection snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Intersection"
        _tip = "Set snapping to the intersection of edges."

        return {'Pixmap': 'Snap_Intersection',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Intersection",
                                              _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Intersection",
                                             _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Intersection, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Intersection')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Intersection"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Intersection_Statusbutton", status)


Gui.addCommand('Draft_Snap_Intersection', Draft_Snap_Intersection())


class Draft_Snap_Parallel(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Parallel tool.

    Set snapping to a direction that is parallel to an edge.
    """

    def __init__(self):
        super(Draft_Snap_Parallel, self).__init__(name=_tr("Parallel snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Parallel"
        _tip = "Set snapping to a direction that is parallel to an edge."

        return {'Pixmap': 'Snap_Parallel',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Parallel", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Parallel", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Parallel, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Parallel')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Parallel"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Parallel_Statusbutton", status)


Gui.addCommand('Draft_Snap_Parallel', Draft_Snap_Parallel())


class Draft_Snap_Endpoint(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Endpoint tool.

    Set snapping to endpoints of an edge.
    """

    def __init__(self):
        super(Draft_Snap_Endpoint, self).__init__(name=_tr("Endpoint snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Endpoint"
        _tip = "Set snapping to endpoints of an edge."

        return {'Pixmap': 'Snap_Endpoint',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Endpoint", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Endpoint", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Endpoint, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Endpoint')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Endpoint"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Endpoint_Statusbutton", status)


Gui.addCommand('Draft_Snap_Endpoint', Draft_Snap_Endpoint())


class Draft_Snap_Angle(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Angle tool.

    Set snapping to points in a circular arc located at multiples
    of 30 and 45 degree angles.
    """

    def __init__(self):
        super(Draft_Snap_Angle, self).__init__(name=_tr("Angle snap (30 and 45 degrees)"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Angles (30 and 45 degrees)"
        _tip = ("Set snapping to points in a circular arc located "
                "at multiples of 30 and 45 degree angles.")

        return {'Pixmap': 'Snap_Angle',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Angle", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Angle", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Angle, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Angle')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Angle"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Angle_Statusbutton", status)


Gui.addCommand('Draft_Snap_Angle', Draft_Snap_Angle())


class Draft_Snap_Center(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Center tool.

    Set snapping to the center of a circular arc.
    """

    def __init__(self):
        super(Draft_Snap_Center, self).__init__(name=_tr("Arc center snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = "Set snapping to the center of a circular arc."

        return {'Pixmap': 'Snap_Center',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Center", "Center"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Center", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Center, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Center')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Center"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Center_Statusbutton", status)


Gui.addCommand('Draft_Snap_Center', Draft_Snap_Center())


class Draft_Snap_Extension(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Extension tool.

    Set snapping to the extension of an edge.
    """

    def __init__(self):
        super(Draft_Snap_Extension, self).__init__(name=_tr("Edge extension snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Extension"
        _tip = "Set snapping to the extension of an edge."

        return {'Pixmap': 'Snap_Extension',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Extension", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Extension", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Extension, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Extension')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Extension"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Extension_Statusbutton", status)


Gui.addCommand('Draft_Snap_Extension', Draft_Snap_Extension())


class Draft_Snap_Near(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Near tool.

    Set snapping to the nearest point of an edge.
    """

    def __init__(self):
        super(Draft_Snap_Near, self).__init__(name=_tr("Near snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = "Set snapping to the nearest point of an edge."

        return {'Pixmap': 'Snap_Near',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Near", "Nearest"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Near", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Near, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Near')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Near"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Near_Statusbutton", status)


Gui.addCommand('Draft_Snap_Near', Draft_Snap_Near())


class Draft_Snap_Ortho(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Ortho tool.

    Set snapping to a direction that is a multiple of 45 degrees
    from a point.
    """

    def __init__(self):
        super(Draft_Snap_Ortho, self).__init__(name=_tr("Orthogonal snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Orthogonal angles (45 degrees)"
        _tip = ("Set snapping to a direction that is a multiple "
                "of 45 degrees from a point.")

        return {'Pixmap': 'Snap_Ortho',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Ortho", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Ortho", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Ortho, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Ortho')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Ortho"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Ortho"+"_Statusbutton", status)


Gui.addCommand('Draft_Snap_Ortho', Draft_Snap_Ortho())


class Draft_Snap_Special(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Special tool.

    Set snapping to the special points defined inside an object.
    """

    def __init__(self):
        super(Draft_Snap_Special, self).__init__(name=_tr("Special point snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Special"
        _tip = "Set snapping to the special points defined inside an object."

        return {'Pixmap': 'Snap_Special',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Special", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Special", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Special, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Special')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Special"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Special_Statusbutton", status)


Gui.addCommand('Draft_Snap_Special', Draft_Snap_Special())


class Draft_Snap_Dimensions(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_Dimensions tool.

    Show temporary linear dimensions when editing an object
    and using other snapping methods.
    """

    def __init__(self):
        super(Draft_Snap_Dimensions, self).__init__(name=_tr("Dimension display"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Show dimensions"
        _tip = ("Show temporary linear dimensions when editing an object "
                "and using other snapping methods.")

        return {'Pixmap': 'Snap_Dimensions',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_Dimensions", _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_Dimensions", _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_Dimensions, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('Dimensions')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_Dimensions"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_Dimensions"+"_Statusbutton", status)


Gui.addCommand('Draft_Snap_Dimensions', Draft_Snap_Dimensions())


class Draft_Snap_WorkingPlane(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_Snap_WorkingPlane tool.

    Restricts snapping to a point in the current working plane.
    If you select a point outside the working plane, for example,
    by using other snapping methods, it will snap to that point's
    projection in the current working plane.
    """

    def __init__(self):
        super(Draft_Snap_WorkingPlane, self).__init__(name=_tr("Working plane snap"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _menu = "Working plane"
        _tip = ("Restricts snapping to a point in the current "
                "working plane.\n"
                "If you select a point outside the working plane, "
                "for example, by using other snapping methods,\n"
                "it will snap to that point's projection "
                "in the current working plane.")

        return {'Pixmap': 'Snap_WorkingPlane',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Snap_WorkingPlane",
                                              _menu),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Snap_WorkingPlane",
                                             _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(Draft_Snap_WorkingPlane, self).Activated()

        if hasattr(Gui, "Snapper"):
            status = Gui.Snapper.toggle_snap('WorkingPlane')
            # change interface consistently
            sync_snap_toolbar_button("Draft_Snap_WorkingPlane"+"_Button", status)
            sync_snap_statusbar_button("Draft_Snap_WorkingPlane_Statusbutton", status)


Gui.addCommand('Draft_Snap_WorkingPlane', Draft_Snap_WorkingPlane())


class ShowSnapBar(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_ShowSnapBar tool.

    Show the snap toolbar if it is hidden.
    """

    def __init__(self):
        super(ShowSnapBar, self).__init__(name=_tr("Show snap toolbar"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = "Show the snap toolbar if it is hidden."

        return {'Pixmap': 'Draft_Snap',
                'MenuText': QT_TRANSLATE_NOOP("Draft_ShowSnapBar",
                                              "Show snap toolbar"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_ShowSnapBar",
                                             _tip)}

    def Activated(self):
        """Execute when the command is called."""
        super(ShowSnapBar, self).Activated()

        if hasattr(Gui, "Snapper"):
            Gui.Snapper.show()


Gui.addCommand('Draft_ShowSnapBar', ShowSnapBar())



# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides lists of commands for the Draft Workbench.

This module returns lists of commands, so that the toolbars
can be initialized by Draft, and by other workbenches.
These commands should be defined in `DraftTools`, and in the individual
modules in `draftguitools`.
"""
## @package init_tools
# \ingroup DRAFT
# \brief This module provides lists of commands for the Draft Workbench.

from PySide.QtCore import QT_TRANSLATE_NOOP

# Comment out commands that aren't ready to be used


def get_draft_drawing_commands():
    """Return the drawing commands list."""
    return ["Draft_Line", "Draft_Wire",  # "Draft_Fillet",
            "Draft_ArcTools",
            "Draft_Circle", "Draft_Ellipse", "Draft_Rectangle",
            "Draft_Polygon", "Draft_BSpline", "Draft_BezierTools",
            "Draft_Point", "Draft_Facebinder",
            "Draft_ShapeString"]


def get_draft_annotation_commands():
    """Return the annotation commands list."""
    return ["Draft_Text", "Draft_Dimension",
            "Draft_Label","Draft_AnnotationStyleEditor"]


def get_draft_array_commands():
    """Return the array commands list."""
    return ["Draft_ArrayTools"]


def get_draft_small_commands():
    """Return a list with only some utilities."""
    return ["Draft_Layer",
            "Draft_WorkingPlaneProxy",
            "Draft_ToggleDisplayMode",
            "Draft_AddToGroup",
            "Draft_SelectGroup",
            "Draft_AddConstruction",
            "Draft_Heal"]


def get_draft_modification_commands():
    """Return the modification commands list."""
    lst = ["Draft_Move", "Draft_Rotate",
           "Draft_Scale", "Draft_Mirror",
           "Draft_Offset", "Draft_Trimex",
           "Draft_Stretch",
           "Separator",
           "Draft_Clone"]
    lst += get_draft_array_commands()
    lst += ["Separator",
            "Draft_Edit", "Draft_SubelementHighlight",
            "Separator",
            "Draft_Join", "Draft_Split",
            "Draft_Upgrade", "Draft_Downgrade",
            "Separator",
            "Draft_WireToBSpline", "Draft_Draft2Sketch",
            "Draft_Slope", "Draft_FlipDimension",
            "Separator",
            "Draft_Shape2DView"]
    return lst


def get_draft_context_commands():
    """Return the context menu commands list."""
    return ["Draft_ApplyStyle", "Draft_ToggleDisplayMode",
            "Draft_AddToGroup", "Draft_SelectGroup",
            "Draft_SelectPlane", "Draft_ShowSnapBar",
            "Draft_ToggleGrid", "Draft_AutoGroup"]


def get_draft_line_commands():
    """Return the line commands list."""
    return ["Draft_UndoLine", "Draft_FinishLine",
            "Draft_CloseLine"]


def get_draft_utility_commands():
    """Return the utility commands list."""
    return ["Draft_Layer", "Draft_Heal", "Draft_FlipDimension",
            "Draft_ToggleConstructionMode",
            "Draft_ToggleContinueMode", "Draft_Edit",
            "Draft_Slope", "Draft_WorkingPlaneProxy",
            "Draft_AddConstruction"]


def get_draft_snap_commands():
    """Return the snapping commands list."""
    return ['Draft_Snap_Lock',
            'Draft_Snap_Endpoint', 'Draft_Snap_Midpoint',
            'Draft_Snap_Center', 'Draft_Snap_Angle',
            'Draft_Snap_Intersection', 'Draft_Snap_Perpendicular',
            'Draft_Snap_Extension', 'Draft_Snap_Parallel',
            'Draft_Snap_Special', 'Draft_Snap_Near',
            'Draft_Snap_Ortho', 'Draft_Snap_Grid',
            'Draft_Snap_WorkingPlane', 'Draft_Snap_Dimensions',
            'Separator', 'Draft_ToggleGrid'
            ]


def init_draft_toolbars(workbench):
    """Initialize the Draft toolbars.

    Parameters
    ----------
    workbench : Gui.Workbench
        The workbench class on which the commands have to be available.
        If called from within the `Initialize` method
        of a workbench class defined inside `InitGui.py`,
        it can be used as `setup_draft_toolbars(self)`.
    """
    workbench.appendToolbar(QT_TRANSLATE_NOOP("Draft",
                                              "Draft creation tools"),
                            get_draft_drawing_commands())
    workbench.appendToolbar(QT_TRANSLATE_NOOP("Draft",
                                              "Draft annotation tools"),
                            get_draft_annotation_commands())
    workbench.appendToolbar(QT_TRANSLATE_NOOP("Draft",
                                              "Draft modification tools"),
                            get_draft_modification_commands())


def init_draft_menus(workbench):
    """Initialize the Draft menus.

    Parameters
    ----------
    workbench : Gui.Workbench
        The workbench class on which the commands have to be available.
        If called from within the `Initialize` method
        of a workbench class defined inside `InitGui.py`,
        it can be used as `setup_draft_menus(self)`.
    """
    workbench.appendMenu(QT_TRANSLATE_NOOP("Draft", "&Drafting"),
                         get_draft_drawing_commands())
    workbench.appendMenu(QT_TRANSLATE_NOOP("Draft", "&Annotation"),
                         get_draft_annotation_commands())
    workbench.appendMenu(QT_TRANSLATE_NOOP("Draft", "&Modification"),
                         get_draft_modification_commands())
    workbench.appendMenu(QT_TRANSLATE_NOOP("Draft", "&Utilities"),
                         get_draft_utility_commands()
                         + get_draft_context_commands())



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""This module provides the view provider code for the base Draft object.
"""
## @package view_base
# \ingroup DRAFT
# \brief This module provides the view provider code for the base Draft object.


import PySide.QtCore as QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App

if App.GuiUp:
    from pivy import coin
    import FreeCADGui as Gui

import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

#import Draft_rc
# from DraftGui import translate
# from DraftGui import displayExternal

# So the resource file doesn't trigger errors from code checkers (flake8)
#True if Draft_rc.__name__ else False

class ViewProviderDraft(object):
    """The base class for Draft view providers.

    Parameters
    ----------
    vobj : a base C++ view provider
        The view provider of the scripted object (`obj.ViewObject`),
        which commonly may be of types `PartGui::ViewProvider2DObjectPython`,
        `PartGui::ViewProviderPython`, or `Gui::ViewProviderPythonFeature`.

        A basic view provider is instantiated during the creation
        of the base C++ object, for example,
        `Part::Part2DObjectPython`, `Part::FeaturePython`,
        or `App::FeaturePython`.

            >>> obj = App.ActiveDocument.addObject('Part::Part2DObjectPython')
            >>> vobj = obj.ViewObject
            >>> ViewProviderDraft(vobj)

        This view provider class instance is stored in the `Proxy` attribute
        of the base view provider.
        ::
            vobj.Proxy = self

    Attributes
    ----------
    Object : the base C++ object
        The scripted document object that is associated
        with this view provider, which commonly may be of types
        `Part::Part2DObjectPython`, `Part::FeaturePython`,
        or `App::FeaturePython`.

    texture : coin.SoTexture2
        A texture that could be added to this object.

    texcoords : coin.SoTextureCoordinatePlane
        The coordinates defining a plane to use for aligning the texture.

    These class attributes are accessible through the `Proxy` object:
    `vobj.Proxy.Object`, `vobj.Proxy.texture`, etc.
    """

    def __init__(self, vobj):
        self.Object = vobj.Object
        self.texture = None
        self.texcoords = None

        vobj.addProperty("App::PropertyEnumeration", "Pattern", "Draft",
                         QT_TRANSLATE_NOOP("App::Property",
                                           "Defines a hatch pattern"))
        vobj.addProperty("App::PropertyFloat", "PatternSize", "Draft",
                         QT_TRANSLATE_NOOP("App::Property",
                                           "Sets the size of the pattern"))
        vobj.Pattern = ["None"] + list(utils.svg_patterns().keys())
        vobj.PatternSize = 1

        # This class is assigned to the Proxy attribute
        vobj.Proxy = self

    def __getstate__(self):
        """Return a tuple of all serializable objects or None.

        When saving the document this view provider object gets stored
        using Python's `json` module.

        Since we have some un-serializable objects (Coin objects) in here
        we must define this method to return a tuple of all serializable
        objects or `None`.

        Override this method to define the serializable objects to return.

        By default it returns `None`.

        Returns
        -------
        None
        """
        return None

    def __setstate__(self, state):
        """Set some internal properties for all restored objects.

        When a document is restored this method is used to set some properties
        for the objects stored with `json`.

        Override this method to define the properties to change for the
        restored serialized objects.

        By default no objects were serialized with `__getstate__`,
        so nothing needs to be done here, and it returns `None`.

        Parameters
        ---------
        state : state
            A serialized object.

        Returns
        -------
        None
        """
        return None

    def attach(self, vobj):
        """Set up the scene sub-graph of the view provider.

        This method should always be defined, even if it does nothing.

        Override this method to set up a custom scene.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.
        """
        self.texture = None
        self.texcoords = None
        self.Object = vobj.Object
        self.onChanged(vobj, "Pattern")
        return

    def updateData(self, obj, prop):
        """This method is run when an object property is changed.

        Override this method to handle the behavior of the view provider
        depending on changes that occur to the real object's properties.

        By default, no property is tested, and it does nothing.

        Parameters
        ----------
        obj : the base C++ object
            The scripted document object that is associated
            with this view provider, which commonly may be of types
            `Part::Part2DObjectPython`, `Part::FeaturePython`,
            or `App::FeaturePython`.

        prop : str
            Name of the property that was modified.
        """
        return

    def getDisplayModes(self, vobj):
        """Return a list of display modes.

        Override this method to return a list of strings with
        display mode styles, such as `'Flat Lines'`, `'Shaded'`,
        `'Wireframe'`, `'Points'`.

        By default it returns an empty list.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        Returns
        -------
        list
            Empty list `[ ]`
        """
        modes = []
        return modes

    def getDefaultDisplayMode(self):
        """Return the default mode defined in getDisplayModes.

        Override this method to return a string with the default display mode.

        By default it returns `'Flat Lines'`.

        Returns
        -------
        str
            `'Flat Lines'`

        """
        return "Flat Lines"

    def setDisplayMode(self, mode):
        """Map the modes defined in attach with those in getDisplayModes.

        This method is optional.

        By default since they have the same names nothing needs to be done,
        and it just returns the input `mode`.

        Parameters
        ----------
        str
            A string defining a display mode such as
            `'Flat Lines'`, `'Shaded'`, `'Wireframe'`, `'Points'`.
        """
        return mode

    def onChanged(self, vobj, prop):
        """This method is run when a view property is changed.

        Override this method to handle the behavior
        of the view provider depending on changes that occur to its properties
        such as line color, line width, point color, point size,
        draw style, shape color, transparency, and others.

        This method  updates the texture and pattern if
        the properties `TextureImage`, `Pattern`, `DiffuseColor`,
        and `PatternSize` change.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        prop : str
            Name of the property that was modified.
        """
        # treatment of patterns and image textures
        if prop in ("TextureImage", "Pattern", "DiffuseColor"):
            if hasattr(self.Object, "Shape"):
                if self.Object.Shape.Faces:
                    path = None
                    if hasattr(vobj, "TextureImage"):
                        if vobj.TextureImage:
                            path = vobj.TextureImage
                    if not path:
                        if hasattr(vobj, "Pattern"):
                            if str(vobj.Pattern) in list(utils.svg_patterns().keys()):
                                path = utils.svg_patterns()[vobj.Pattern][1]
                            else:
                                path = "None"
                    if path and vobj.RootNode:
                        if vobj.RootNode.getChildren().getLength() > 2:
                            if vobj.RootNode.getChild(2).getChildren().getLength() > 0:
                                if vobj.RootNode.getChild(2).getChild(0).getChildren().getLength() > 2:
                                    r = vobj.RootNode.getChild(2).getChild(0).getChild(2)
                                    i = QtCore.QFileInfo(path)
                                    if self.texture:
                                        r.removeChild(self.texture)
                                        self.texture = None
                                    if self.texcoords:
                                        r.removeChild(self.texcoords)
                                        self.texcoords = None
                                    if i.exists():
                                        size = None
                                        if ".SVG" in path.upper():
                                            size = utils.get_param("HatchPatternResolution", 128)
                                            if not size:
                                                size = 128
                                        im = gui_utils.load_texture(path, size)
                                        if im:
                                            self.texture = coin.SoTexture2()
                                            self.texture.image = im
                                            r.insertChild(self.texture, 1)
                                            if size:
                                                s = 1
                                                if hasattr(vobj, "PatternSize"):
                                                    if vobj.PatternSize:
                                                        s = vobj.PatternSize
                                                self.texcoords = coin.SoTextureCoordinatePlane()
                                                self.texcoords.directionS.setValue(s, 0, 0)
                                                self.texcoords.directionT.setValue(0, s, 0)
                                                r.insertChild(self.texcoords, 2)
        elif prop == "PatternSize":
            if hasattr(self, "texcoords"):
                if self.texcoords:
                    s = 1
                    if vobj.PatternSize:
                        s = vobj.PatternSize
                    vS = App.Vector(self.texcoords.directionS.getValue().getValue())
                    vT = App.Vector(self.texcoords.directionT.getValue().getValue())
                    vS.Length = s
                    vT.Length = s
                    self.texcoords.directionS.setValue(vS.x, vS.y, vS.z)
                    self.texcoords.directionT.setValue(vT.x, vT.y, vT.z)
        return

    def _update_pattern_size(self, vobj):
        """Update the pattern size. Helper method in onChanged."""
        if hasattr(self, "texcoords") and self.texcoords:
            s = 1
            if vobj.PatternSize:
                s = vobj.PatternSize
            vS = App.Vector(self.texcoords.directionS.getValue().getValue())
            vT = App.Vector(self.texcoords.directionT.getValue().getValue())
            vS.Length = s
            vT.Length = s
            self.texcoords.directionS.setValue(vS.x, vS.y, vS.z)
            self.texcoords.directionT.setValue(vT.x, vT.y, vT.z)

    def execute(self, vobj):
        """This method is run when the object is created or recomputed.

        Override this method to produce effects when the object
        is newly created, and whenever the document is recomputed.

        By default it does nothing.

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.
        """
        return

    def setEdit(self, vobj, mode=0):
        """Enter edit mode of the object.

        Override this method to define a custom command to run when entering
        the edit mode of the object in the tree view.
        It must return `True` to successfully enter edit mode.
        If the conditions to edit are not met, it should return `False`,
        in which case the edit mode is not started.

        By default it runs the `Draft_Edit` GuiCommand.
        ::
            Gui.runCommand('Draft_Edit')

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        mode : int, optional
            It defaults to 0, in which case
            it runs the `Draft_Edit` GuiCommand.
            It indicates the type of edit in the underlying C++ code.

        Returns
        -------
        bool
            It is `True` if `mode` is 0, and `Draft_Edit` ran successfully.
            It is `False` otherwise.
        """
        if mode == 0 and App.GuiUp: #remove guard after splitting every viewprovider
            Gui.runCommand("Draft_Edit")
            return True
        return False

    def unsetEdit(self, vobj, mode=0):
        """Terminate the edit mode of the object.

        Override this method to define a custom command to run when
        terminating the edit mode of the object in the tree view.

        It should return `True` to indicate that the method already
        cleaned up everything and there is no need to call
        the `usetEdit` method of the base class.
        It should return `False` to indicate that cleanup
        is still required, so the `unsetEdit` method of the base class
        is invoked to do the rest.

        By default it runs the `finish` method of the active
        Draft GuiCommand, and closes the task panel.
        ::
            App.activeDraftCommand.finish()
            Gui.Control.closeDialog()

        Parameters
        ----------
        vobj : the view provider of the scripted object.
            This is `obj.ViewObject`.

        mode : int, optional
            It defaults to 0. It is not used.
            It indicates the type of edit in the underlying C++ code.

        Returns
        -------
        bool
            This method always returns `False` so it passes
            control to the base class to finish the edit mode.
        """
        if App.activeDraftCommand:
            App.activeDraftCommand.finish()
        if App.GuiUp: # remove guard after splitting every viewprovider
            Gui.Control.closeDialog()
        return False

    def getIcon(self):
        """Return the path to the icon used by the view provider.

        The path can be a full path in the system, or a relative path
        inside the compiled resource file.
        It can also be a string that defines the icon in XPM format.

        Override this method to provide a specific icon
        for the object in the tree view.

        By default it returns the path to the `Draft_Draft.svg` icon.

        Returns
        -------
        str
            `':/icons/Draft_Draft.svg'`
        """
        tp = self.Object.Proxy.Type
        if tp in ('Line', 'Wire', 'Polyline'):
            return ":/icons/Draft_N-Linear.svg"
        elif tp in ('Rectangle', 'Polygon'):
            return ":/icons/Draft_N-Polygon.svg"
        elif tp in ('Circle', 'Ellipse', 'BSpline', 'BezCurve', 'Fillet'):
            return ":/icons/Draft_N-Curve.svg"
        elif tp in ("ShapeString"):
            return ":/icons/Draft_ShapeString_tree.svg"
        else:
            return ":/icons/Draft_Draft.svg"
        return ":/icons/Draft_Draft.svg"

    def claimChildren(self):
        """Return objects that will be placed under it in the tree view.

        Override this method to return a list with objects
        that will appear under this object in the tree view.
        That is, this object becomes the `parent`,
        and all those under it are the `children`.

        By default the returned list is composed of objects from
        `Object.Base`, `Object.Objects`, `Object.Components`,
        and `Object.Group`, if they exist.

        Returns
        -------
        list
            List of objects.
        """
        objs = []
        if hasattr(self.Object, "Base"):
            objs.append(self.Object.Base)
        if hasattr(self.Object, "Objects"):
            objs.extend(self.Object.Objects)
        if hasattr(self.Object, "Components"):
            objs.extend(self.Object.Components)
        if hasattr(self.Object, "Group"):
            objs.extend(self.Object.Group)
        return objs


_ViewProviderDraft = ViewProviderDraft


class ViewProviderDraftAlt(ViewProviderDraft):
    """A view provider that doesn't absorb its base object in the tree view.

    The `claimChildren` method is overridden to return an empty list.
    """

    def __init__(self, vobj):
        super(ViewProviderDraftAlt, self).__init__(vobj)

    def claimChildren(self):
        objs = []
        return objs


_ViewProviderDraftAlt = ViewProviderDraftAlt


class ViewProviderDraftPart(ViewProviderDraftAlt):
    """A view provider that displays a Part icon instead of a Draft icon.

    The `getIcon` method is overridden to provide `Tree_Part.svg`.
    """

    def __init__(self, vobj):
        super(ViewProviderDraftPart, self).__init__(vobj)

    def getIcon(self):
        return ":/icons/Tree_Part.svg"


_ViewProviderDraftPart = ViewProviderDraftPart